/*
VODKA-INDUCED ENTERTAINMENT ADVANCED SOUND MANAGER v0.9.1
COMPATIBILITY LAYER FOR WOLF4SDL v1.6 - HEADER FILE
BY GERARD 'ALUMIUN' WATSON
*/

#ifndef _ASMCOMP_H_
#define _ASMCOMP_H_

#include "vieasm.h"

#define TickBase        70      // 70Hz per tick - used as a base for timer 0

#define GetTimeCount()  ((SDL_GetTicks()*TickBase)/1000)

inline void Delay(int wolfticks)
{
    if (wolfticks>0) 
        SDL_Delay(wolfticks * 1000 / TickBase);
}

#define SD_SoundPlaying() ASM_SoundPlaying()
#define UNCACHEAUDIOCHUNK(chunk) ASM_Uncache(ASM_Audiosegs[chunk]);
#define SD_WaitSoundDone()
#define SD_PrepareSound(which) int nullps = which
#define SD_AbortIfMusOff if (!SD_MusIsOn()) return
#define SD_AbortIfSndOff if (!SD_SndIsOn()) return
#define SD_StopDigitized() SD_StopSound()
#define SD_ContinueMusic(chunk, startoffs) SD_StartMusic(chunk)
#define SD_ChangeVolume(snd, mus) ASM_ChangeVolume(snd, mus)
#define SD_GetVolume(snd, mus) ASM_ReturnVolume(snd, mus)
#define SD_Reverse(reverse) ASM_ReverseStereo(reverse)
#define SD_StopChannel(channel) ASM_StopChannel(channel)
#define SD_FadeOutChannel(channel, ms) ASM_FadeOutChannel(channel, ms)

extern sample *ASM_Audiosegs;

// These enums are now treated differently. Because there are no seperate devices,
// any value other than sxm_Off will turn the device on, and SDSMode is no longer used.

typedef enum
{
    sdm_Off,sdm_AdLib
} SDMode;

typedef enum
{
    smm_Off,smm_AdLib
} SMMode;

typedef struct
{
    int valid;
    fixed globalsoundx, globalsoundy;
    int maxTileDist;
} globalsoundpos;

extern globalsoundpos channelSoundPos[];

extern  SDMode          SoundMode;
extern  SMMode          MusicMode;

// Functions

extern  void    SD_InitAudioSegs(void);
extern  void    SD_Startup(void), SD_Shutdown(void);

extern  void    SD_PositionSound(int leftvol, int rightvol, int volume);
extern  int     SD_PlaySound(int sound);
extern  int     SD_PlaySoundEx(ASM_SoundPlay_t soundPlay);
extern  int     SD_PlaySoundVolume(int sound, int volume);
extern  void    SD_SetPosition(int channel, int leftpos, int rightpos,
    int distance);
extern  void    SD_StopSound(void);

extern  void    SD_StartMusic(int chunk);
extern  void    SD_ContinueMusic(int chunk, int startoffs);
extern  void    SD_MusicOn(void), SD_FadeOutMusic(int fade);
extern  int     SD_MusicOff(void);

extern  bool    SD_MusicPlaying(void);
extern  bool    SD_SetSoundMode(SDMode mode);
extern  bool    SD_SetMusicMode(SMMode mode);

extern  int     SD_PlayDigitized(ASM_SoundPlay_t soundPlay);
extern  void    CA_LoadAllSounds(void);

extern  bool    SD_ChannelPlaying(int channel);
extern  int     SD_MaxVolume(void);
extern  void    SD_SetChannelVol(int channel, voltype vol);

#endif  // _ASMCOMP_H_
