/*************************************************************************
** Wolf3D Legacy
** Copyright (C) 2012 by LinuxWolf - Team RayCAST
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**************************************************************************
** LinuxWolf Library for Wolf3D Legacy
*************************************************************************/

#ifndef LWLIB_FS_H
#define LWLIB_FS_H

#include <stdlib.h>
#include <stdint.h>
#include "lw_misc.h"

#define lwlib_Fs_FileMode(name) LWLIB_FS_FILEMODE_ ## name

#define lwlib_Fs_CommitBlock(fs, block, len, varType, desc) \
do { \
	if ((fs) != NULL) \
	{ \
		(fs)->commit(block, len, varType, desc); \
	} \
} while (0)

#define lwlib_Fs_CommitVar(fs, var, varType, desc) \
	lwlib_Fs_CommitBlock(fs, var, sizeof(*(var)), varType, desc)

namespace lwlib
{
	namespace Fs_FileMode
	{
		enum e
		{
			Read,
			Write,
			ReadText,
			WriteText,
			Last = WriteText,
		};
	}

	namespace Fs_VarType
	{
		enum e
		{
			Int,
			Short,
			UChar,
			Double,
			Bool,
			Raw,
			String,
			Informative,
		};
	}

	lwlib_EnumStrExtern(Fs_FileMode);

	class Fs
	{
		struct
		{
			FILE *fp;
			Fs_FileMode::e fileMode;
			lwlib_Byte_t *data;
			int length;
		} priv;

		void consumeLine(lwlib_String_t s);
		
	public:
		Fs(const char *fileName, Fs_FileMode::e fileMode);

		~Fs();

		void commit(void *buf, int count, Fs_VarType::e varType, 
			const char *desc);

		bool loadEntireFile();

		lwlib_Byte_t *getData();

		int getLength();

		Fs_FileMode::e getFileMode();
	};
}

#endif
