/*************************************************************************
** Wolf3D Legacy
** Copyright (C) 2012 by LinuxWolf - Team RayCAST
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**************************************************************************
** LinuxWolf Library for Wolf3D Legacy
*************************************************************************/

#ifndef LWLIB_GL_H
#define LWLIB_GL_H

#ifdef LEGACYENGINE_OPENGL
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glext.h>
#endif
#include <SDL.h>
#include "lw_misc.h"
#include "lw_vec.h"
#include "lw_sdlglemu.h"

namespace lwlib
{
	typedef struct Gl_Screen_s
	{
		int width;
		int height;
		int bits;
		bool fullscreen;
		bool useDoubleBuffering;
		SDL_Surface *screenSurf;
		bool isDoubleBuffered;
	#ifndef SDL_TRUE_COLOR
		SDL_Color *palette;
	#endif
	} Gl_Screen_t;

#ifdef LEGACYENGINE_OPENGL
	typedef vec_t(GLuint) Gl_TexIdSeq_t;
#endif

	class Gl
	{
		struct
		{
#ifdef LEGACYENGINE_OPENGL
			Gl_TexIdSeq_t texIdSeq;
#endif
			SDLGlEmu_t sdlGlEmu;
			int screenWidth;
			int screenHeight;
			int screenPitch;
			bool isDoubleBuffered;
		} priv;

	public:
		Gl(bool sdlEmu);

		~Gl();

		void setVideoMode(Gl_Screen_t *screen);

		void init();

		void clearScreen();

		void updateScreen();

		int uploadTex(lwlib_Byte_t *data, int w, int h, int bpp, 
			SDL_Color *palette);

		void deleteTex(int texId);

		void lockScreen();

		void unlockScreen();

		void drawScaledImage(int texId, int ix, int iy, int iw, int ih, 
			int w, int h, int cx, int cy, int sw, int sh, bool transparent);

		void setCurColor(Point4b colorRgba);

		void resetCurColor();

		void drawBox(int x, int y, int w, int h, Point4b colorRgba, 
			bool lines);

		void drawBoxArrow(int x, int y, int w, int h, int r, 
			Point4b colorRgba, double dir);

		void pushMatrix();

		void popMatrix();

		void translateMatrix(double tx, double ty, double tz);

		void drawLine(int x1, int y1, int x2, int y2,
			Point4b colorRgba);

		int getWidth();

		int getHeight();

		Point2i getDims();
	};

#ifndef LWLIB_GL_PTRTYPE
#define LWLIB_GL_PTRTYPE
	typedef Gl *Gl_t;
#endif
}

#endif
