/*************************************************************************
** Wolf3D Legacy
** Copyright (C) 2012 by LinuxWolf - Team RayCAST
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**************************************************************************
** LinuxWolf Library for Wolf3D Legacy
*************************************************************************/

#include <sys/types.h>
#if defined _WIN32
	#include <io.h>
#else
	#include <sys/uio.h>
	#include <unistd.h>
#endif
#include "lw_maptool.h"
#include "lw_timer.h"
#include "lw_command.h"
#include "wl_def.h"
using namespace lwlib;

//#define LWLIB_MAPTOOL_EDITORID "LWMTv1.0"
#define LWLIB_MAPTOOL_EDITORID "CE1.27:3P:FF:Mapfile was edited with ChaosEdit (www.chaos-software.de.vu)"
#define LWLIB_MAPTOOL_RED 4
#define LWLIB_MAPTOOL_ORANGE 6
#define LWLIB_MAPTOOL_BLUE 7
#define LWLIB_MAPTOOL_WHITE 8
#define LWLIB_MAPTOOL_OTHER 9

#define lwlib_MapTool_FileFormat(name) LWLIB_MAPTOOL_FILEFORMAT_ ## name

#ifdef MAPSPOT
#undef MAPSPOT
#endif
#define MAPSPOT(x, y, plane) map->planes[plane][(y) * MAPSIZE + (x)]
#define CURMAPSPOT(plane) MAPSPOT(lwlib_X(priv.tilePos), \
	lwlib_Y(priv.tilePos), plane)

#define CURMAPSPOT_ISSOLID() MAPSPOT_ISSOLID(lwlib_X(priv.tilePos), \
	lwlib_Y(priv.tilePos))

#ifdef SDL_TRUE_COLOR
#define BOX_USE_LINES false
#else
#define BOX_USE_LINES true
#endif

lwlib_EnumStr(
	MapTool_FileFormat,
	"Game",
	"Raw",
);

lwlib_EnumStr(
	MapTool_YankMode,
	"Wall",
	"Actor",
	"Floor",
	"WallPatch",
	"Picker",
	"TileBox",
);

void MapTool::map_UpdateEdit(Edit *edit)
{
	int i;
	int type;
	word maptile;
	ED_Spawn_t *spawn;
	MapTool_Map_t *map;
	lwlib_String_t tmpStr;

	map = &priv.maps[priv.mapIndex];

	edit->addItem_short("width", (short *)&map->width);
	edit->addItem_short("height", (short *)&map->height);
	edit->addItem_string("name", map->name, sizeof(map->name));
	
	edit->addItem_int("mapX", &lwlib_X(priv.tilePos));
	edit->addItem_int("mapY", &lwlib_Y(priv.tilePos));

	for (i = 0; i < 3; i++)
	{
		lwlib_Snprintf(tmpStr, "tile[%d]", i);
		edit->addItem_short(tmpStr, (short *)&CURMAPSPOT(i));
	}
	edit->addItem_uchar("floor", &((unsigned char *)&CURMAPSPOT(2))[0]);
	edit->addItem_uchar("ceiling", &((unsigned char *)&CURMAPSPOT(2))[1]);

	maptile = CURMAPSPOT(1);

	type = ED_ActorStatInfo(ed, maptile);
	if (type != -1)
	{
		edit->addItem_string("statInfo.name", 
			ED_STATINFO[type].name, sizeof(ED_STATINFO[type].name));
	}

	spawn = ED_ActorSpawn(ed, maptile);
	if (spawn != NULL)
	{
		edit->addItem_string("spawn.enemyName", 
			spawn->enemyName, sizeof(spawn->enemyName));
	}

	edit->addItem_int("secretLevelIndex", &map->secretLevelIndex);
}

int MapTool::map_ComputeSecretLevel(MapTool_Map_t *map)
{
	int x, y;

	for (x = 0; x < MAPSIZE; x++)
	{
		for (y = 0; y < MAPSIZE; y++)
		{
			if (MAPSPOT(x, y, 0) == 0 && 
				MAPSPOT(x, y, 1) == ENDLEVELTILE)
			{
				return MAPSPOT(0, 3, 0);
			}
		}
	}

	return 0;
}

void MapTool::loadMaps_FileFormatGame()
{
	int i, j;
	maptype *mapheader;
	MapTool_Map_t *map;

	memset(priv.maps, 0, sizeof(priv.maps));

	for (i = 0; i < NUMMAPS; i++)
	{
		map = &priv.maps[i];
		mapheader = mapheaderseg[i];
		if (mapheader == NULL)
		{
			continue;
		}

		map->width = mapheader->width;
		map->height = mapheader->height;
		lwlib_Copy(map->name, mapheader->name);

		if (!(mapheader->width == 64 && mapheader->height == 64))
		{
			continue;
		}

		CA_CacheMap(i);
		for (j=0; j < MAPPLANES; j++)
		{
			memcpy(map->planes[j], mapsegs[j], maparea * 2);
		}

		map->secretLevelIndex = map_ComputeSecretLevel(map);
	}
}

void MapTool::loadMaps_FileFormatRaw()
{
	int i, j;
	FILE *fp;
	MapTool_Map_t *map;
	char tmpStr[32];

	fp = fopen(priv.rawFileName, "rb");
	if (fp == NULL)
	{
		return;
	}

	for (i = 0; i < NUMMAPS; i++)
	{
		map = &priv.maps[i];
		fread(tmpStr, 32, 1, fp);
		fread(&map->width, sizeof(map->width), 1, fp);
		fread(&map->height, sizeof(map->height), 1, fp);
		fread(map->name, sizeof(map->name), 1, fp);
		for (j = 0; j < MAPPLANES; j++)
		{
			fread(tmpStr, 32, 1, fp);
			fread(map->planes[j], sizeof(map->planes[j]), 1, fp);
		}
	}

	fclose(fp);
}

void MapTool::loadMaps()
{

	switch (priv.fileFormat)
	{
	case MapTool_FileFormat::Game:
		loadMaps_FileFormatGame();
		break;
	case MapTool_FileFormat::Raw:
		loadMaps_FileFormatRaw();
		break;
	}
}

int MapTool::saveMapHead()
{
	int i;
	int off;
	int handle;
	char fname[13];
	int32_t length;
	mapfiletype tinf;
	MapTool_Map_t *map;

	strcpy(fname,mheadname);
	strcat(fname,extension);

	handle = open(fname, O_RDWR | O_BINARY);
	if (handle == -1)
		CA_CannotOpen(fname);

	lwlib_Zero(tinf);
	tinf.RLEWtag = RLEWtag;

	off = strlen(LWLIB_MAPTOOL_EDITORID);

	for (i = 0; i < NUMMAPS; i++)
	{
		map = &priv.maps[i];
		if (!(map->width == 64 && map->height == 64))
		{
			tinf.headeroffsets[i] = -1;
			continue;
		}

		tinf.headeroffsets[i] = off;
		off += sizeof(maptype);
	}

	length = NUMMAPS * 4 + 2;
	write(handle, &tinf, length);
	
	close(handle);

	return off;
}

lwlib_Byte_t *MapTool::mapPlaneOnDisk(int mapNum, int planeNum,
	int *onDiskSize)
{
	int expanded, rlew_size, compressed;
	byte *planeseg, *rlew_planeseg;
	MapTool_Map_t *map;

	map = &priv.maps[mapNum];

	expanded = maparea * 2;
	rlew_planeseg = lwlib_CallocMany(byte, expanded + 2);
	rlew_size = CA_RLEWCompress(map->planes[planeNum], expanded,
		((word *)rlew_planeseg) + 1, RLEWtag);
	*((word *)rlew_planeseg) = (word)expanded;

	expanded = rlew_size + 2; 
	planeseg = lwlib_CallocMany(byte, (expanded * 3 / 2) + 2);
	compressed = CAL_CarmackCompress((word *)rlew_planeseg, expanded,
		((word *)planeseg) + 1);
	*((word *)planeseg) = (word)expanded;

	free(rlew_planeseg);

	if (onDiskSize != NULL)
	{
		*onDiskSize = compressed + 2;
	}
	return planeseg;
}

int MapTool::mapPlaneSizeOnDisk(int mapNum, int planeNum)
{
	int onDiskSize;
	byte *planeseg;

	planeseg = mapPlaneOnDisk(mapNum, planeNum, &onDiskSize);
	free(planeseg);

	return onDiskSize;
}

void MapTool::saveMapHeaders(int handle, int off)
{
	int i, j;
	int size;
	MapTool_Map_t *map;
	maptype mapheader;

	for (i = 0; i < NUMMAPS; i++)
	{
		map = &priv.maps[i];
		if (!(map->width == 64 && map->height == 64))
		{
			continue;
		}

		lwlib_Zero(mapheader);
		for (j = 0; j < 3; j++)
		{
			size = mapPlaneSizeOnDisk(i, j);
			mapheader.planestart[j] = off;
			mapheader.planelength[j] = size;
			off += size;
		}
		mapheader.width = mapheader.height = 64;
		lwlib_Copy(mapheader.name, map->name);

		write(handle, &mapheader, sizeof(mapheader));
	}
}

void MapTool::saveMapPlanes(int handle)
{
	int i, j;
	int onDiskSize;
	byte *planeseg;
	MapTool_Map_t *map;

	for (i = 0; i < NUMMAPS; i++)
	{
		map = &priv.maps[i];
		if (!(map->width == 64 && map->height == 64))
		{
			continue;
		}

		for (j = 0; j < 3; j++)
		{
			planeseg = mapPlaneOnDisk(i, j, &onDiskSize);
			write(handle, planeseg, onDiskSize);
			free(planeseg);
		}
	}
}

void MapTool::saveMaps_FileFormatGame()
{
	int off;
	int handle;
	char fname[13];

	off = saveMapHead();

	strcpy(fname, "gamemaps.");
	strcat(fname, extension);

	handle = open(fname, O_RDWR | O_BINARY);
	if (handle == -1)
	{
		CA_CannotOpen(fname);
	}

	write(handle, LWLIB_MAPTOOL_EDITORID, strlen(LWLIB_MAPTOOL_EDITORID));

	saveMapHeaders(handle, off);
	saveMapPlanes(handle);

	close(handle);
}

void MapTool::saveMaps_FileFormatRaw()
{
	int i, j;
	FILE *fp;
	MapTool_Map_t *map;
	lwlib_String_t tmpStr;

	fp = fopen(priv.rawFileName, "wb");
	if (fp == NULL)
	{
		return;
	}

	for (i = 0; i < NUMMAPS; i++)
	{
		map = &priv.maps[i];
		lwlib_Zero(tmpStr);
		fwrite(lwlib_VaString(tmpStr, "MAP%d", i), 32, 1, fp);
		fwrite(&map->width, sizeof(map->width), 1, fp);
		fwrite(&map->height, sizeof(map->height), 1, fp);
		fwrite(map->name, sizeof(map->name), 1, fp);
		for (j = 0; j < MAPPLANES; j++)
		{
			#if 0
			if (j == 2)
			{
				int k;
				unsigned top, bot;

				for (k = 0; k < maparea; k++)
				{
					top = map->planes[j][k] >> 8;
					bot = map->planes[j][k] & 0xff;
					top += (top != 0 ? 3 : 0);
					bot += (bot != 0 ? 3 : 0);
					map->planes[j][k] = (top << 8) + bot;
				}
			}
			#endif
			lwlib_Zero(tmpStr);
			fwrite(lwlib_VaString(tmpStr, "MAP%d,PLANE%d", i, j), 32, 1, fp);
			fwrite(map->planes[j], sizeof(map->planes[j]), 1, fp);
		}
	}

	fclose(fp);
}

void MapTool::saveMaps()
{

	switch (priv.fileFormat)
	{
	case MapTool_FileFormat::Game:
		saveMaps_FileFormatGame();
		break;
	case MapTool_FileFormat::Raw:
		saveMaps_FileFormatRaw();
		break;
	}
}

void MapTool::tilePosMotion()
{
	int i;
	bool panning;
	Point2i mouseMove;

	panning = priv.commandSet.panView->pending();

	mouseMove = priv.mouse->getMove();
	for (i = 0; i < 2; i++)
	{
		if (lwlib_C(mouseMove, i) == 0)
		{
			continue;
		}

		lwlib_C(priv.tileMove, i) += (double)lwlib_C(mouseMove, i) *
			priv.tileMoveSensitivity;
		while (lwlib_C(priv.tileMove, i) >= 1.0)
		{
			lwlib_C(priv.tilePos, i)++;
			if (panning)
			{
				lwlib_C(priv.pan, i) -= priv.grid;
			}

			lwlib_C(priv.tileMove, i) -= 1.0;
		}
		while (lwlib_C(priv.tileMove, i) < 0.0)
		{
			lwlib_C(priv.tilePos, i)--;
			if (panning)
			{
				lwlib_C(priv.pan, i) += priv.grid;
			}

			lwlib_C(priv.tileMove, i) += 1.0;
		}
	}
}

void MapTool::panMotion()
{
	int i;
	Point2i glDims;

	glDims = priv.gl->getDims();

	for (i = 0; i < 2; i++)
	{
		if (lwlib_C(priv.tilePos, i) * priv.grid < 
			-lwlib_C(priv.pan, i))
		{
			lwlib_C(priv.pan, i) = -lwlib_C(priv.tilePos, i) * 
				priv.grid;
		}
		else if ((lwlib_C(priv.tilePos, i) + 1) * priv.grid > 
			-lwlib_C(priv.pan, i) + lwlib_C(glDims, i))
		{
			lwlib_C(priv.pan, i) = lwlib_C(glDims, i) - 
				((lwlib_C(priv.tilePos, i) + 1) * priv.grid);
		}
	}

	if (priv.commandSet.zoomIn->pending())
	{
		priv.grid += 16;
		if (priv.grid > 128)
		{
			priv.grid = 128;
		}
	}

	if (priv.commandSet.zoomOut->pending())
	{
		priv.grid -= 16;
		if (priv.grid <= 0)
		{
			priv.grid = 16;
		}
	}
}

void MapTool::setAreaTile(Point2i pos)
{
	int i;
	int x, y;
	Point2i neigh;
	MapTool_Map_t *map;

	map = &priv.maps[priv.mapIndex];

	x = lwlib_X(pos);
	y = lwlib_Y(pos);

	if (MAPSPOT_ISSOLID(x, y) || 
		priv.areaTileVisit[y * MAPSIZE + x])
	{
		return;
	}
	priv.areaTileVisit[y * MAPSIZE + x] = true;

	if (MAPSPOT(x, y, 0) == 0 || MAPSPOT(x, y, 0) >= AREATILE)
	{
		MAPSPOT(x, y, 0) = priv.areaCode + AREATILE;
	}

	for (i = 0; i < 4; i++)
	{
		neigh = pos;
		lwlib_C(neigh, i / 2) += (i % 2) * 2 - 1;
		setAreaTile(neigh);
	}
}

void MapTool::removeSolidWall()
{
	int x, y;
	MapTool_Map_t *map;

	map = &priv.maps[priv.mapIndex];

	x = lwlib_X(priv.tilePos);
	y = lwlib_Y(priv.tilePos);
	if (!MAPSPOT_ISSOLID(x - 1, y))
	{
		CURMAPSPOT(0) = MAPSPOT(x - 1, y, 0);
		CURMAPSPOT(1) = 0;
		CURMAPSPOT(2) = MAPSPOT(x - 1, y, 2);
	}
	if (!MAPSPOT_ISSOLID(x + 1, y))
	{
		CURMAPSPOT(0) = MAPSPOT(x + 1, y, 0);
		CURMAPSPOT(1) = 0;
		CURMAPSPOT(2) = MAPSPOT(x + 1, y, 2);
	}
	if (!MAPSPOT_ISSOLID(x, y - 1))
	{
		CURMAPSPOT(0) = MAPSPOT(x, y - 1, 0);
		CURMAPSPOT(1) = 0;
		CURMAPSPOT(2) = MAPSPOT(x, y - 1, 2);
	}
	if (!MAPSPOT_ISSOLID(x, y + 1))
	{
		CURMAPSPOT(0) = MAPSPOT(x, y + 1, 0);
		CURMAPSPOT(1) = 0;
		CURMAPSPOT(2) = MAPSPOT(x, y + 1, 2);
	}
}

void MapTool::setYankMode(MapTool_YankMode::e yankMode)
{
	priv.yank.mode = yankMode;
	if (yankMode != MapTool_YankMode::TileBox)
	{
		priv.yank.tileBoxForming = false;
		lwlib_Zero(priv.yank.tileBox);
	}
}

void MapTool::yankTile()
{
	int tile;
	int floor, ceiling;
	Point2i pos;
	Vector2i span;
	MapTool_Map_t *map;
	lwlib_MapTool_Yank_t *yank;

	map = &priv.maps[priv.mapIndex];
	yank = &priv.yank;

	if (yank->mode == MapTool_YankMode::TileBox)
	{
		if (yank->tileBoxForming)
		{
			lwlib_box_upper(yank->tileBox) = priv.tilePos;
			yank->tileBox = box_normalize(yank->tileBox);
			yank->tileBoxForming = false;
		}

		pos = lwlib_box_lower(yank->tileBox);
		span = box_span(yank->tileBox);

		CA_CopyPlanes(yank->planes, 0, 0, map->planes, lwlib_X(pos), 
			lwlib_Y(pos), lwlib_X(span) + 1, lwlib_Y(span) + 1);
	}
	else if (CURMAPSPOT_ISSOLID())
	{
		tile = CURMAPSPOT(1);
		if (tile >= ED_GetTileInfo(ed)->firstWallPatchActor && 
			tile <= ED_GetTileInfo(ed)->lastWallPatchActor)
		{
			setYankMode(MapTool_YankMode::WallPatch);
			priv.yank.wall = CURMAPSPOT(0);
			priv.yank.wallPatch = tile;
		}
		else
		{
			setYankMode(MapTool_YankMode::Wall);
			priv.yank.wall = CURMAPSPOT(0);
		}
	}
	else if (CURMAPSPOT(1) != 0)
	{
		setYankMode(MapTool_YankMode::Actor);
		priv.yank.actor = CURMAPSPOT(1);
		priv.yank.wall = CURMAPSPOT(0);
	}
	else
	{
		floor = CURMAPSPOT(2) & 0xff;
		ceiling = (CURMAPSPOT(2) >> 8) & 0xff;
		
		if (ceiling != 0 || floor != 0)
		{
			setYankMode(MapTool_YankMode::FloorCeil);
			priv.yank.ceiling = ceiling;
			priv.yank.floor = floor;
			priv.yank.wall = CURMAPSPOT(0);
		}
	}
}

void MapTool::insertPicked()
{
	int x, y;
	int tile;
	int vertical;
	MapTool_Map_t *map;
	Picker_ObjType::e pickerObjType;

	map = &priv.maps[priv.mapIndex];

	tile = priv.picker->tileSelector();
	pickerObjType = priv.picker->getObjType();

	switch (pickerObjType)
	{
	case Picker_ObjType::Wall:
		CURMAPSPOT(0) = tile;
		CURMAPSPOT(1) = 0;
		CURMAPSPOT(2) = 0;
		break;
	case Picker_ObjType::Floor:
	case Picker_ObjType::Ceiling:
		if (CURMAPSPOT_ISSOLID())
		{
			removeSolidWall();
		}

		if (pickerObjType == Picker_ObjType::Floor)
		{
			CURMAPSPOT(2) = (CURMAPSPOT(2) & 0xff00) + tile;
		}
		else
		{
			CURMAPSPOT(2) = (CURMAPSPOT(2) & 0xff) + (tile << 8);
		}
		break;
	case Picker_ObjType::WallPatch:
		if (CURMAPSPOT_ISSOLID())
		{
			CURMAPSPOT(1) = tile + 
				ED_GetTileInfo(ed)->firstWallPatchActor;
			CURMAPSPOT(2) = 0;
		}
		break;
	case Picker_ObjType::StatActor:
		if (!CURMAPSPOT_ISSOLID())
		{
			CURMAPSPOT(1) = ED_STATINFO[tile].actor;
		}
		break;
	case Picker_ObjType::EnemyActor:
		if (!CURMAPSPOT_ISSOLID())
		{
			CURMAPSPOT(1) = vec_at(ED_Spawns(ed), tile).actor;
		}
		break;
	case Picker_ObjType::Door:
		x = lwlib_X(priv.tilePos);
		y = lwlib_Y(priv.tilePos);

		vertical = -1;
		if (MAPSPOT_ISSOLID(x + 1, y) && MAPSPOT_ISSOLID(x - 1, y))
		{
			vertical = 1;
		}
		else if (MAPSPOT_ISSOLID(x, y + 1) && 
			MAPSPOT_ISSOLID(x, y - 1))
		{
			vertical = 0;
		}

		if (vertical != -1)
		{
			if (CURMAPSPOT_ISSOLID())
			{
				removeSolidWall();
			}

			CURMAPSPOT(0) = (tile * 2) + DOORTILE_FIRST + vertical;
			CURMAPSPOT(1) = 0;
		}
		break;
	case Picker_ObjType::Area:
		if (!CURMAPSPOT_ISSOLID())
		{
			CURMAPSPOT(0) = tile + AREATILE;
		}
		break;
	default:
		break;
	}
}

void MapTool::pasteTileBox()
{
	Point2i pos;
	Vector2i span;
	MapTool_Map_t *map;
	lwlib_MapTool_Yank_t *yank;

	map = &priv.maps[priv.mapIndex];
	yank = &priv.yank;

	if (yank->tileBoxForming)
	{
		return;
	}

	pos = priv.tilePos;
	span = box_span(yank->tileBox);

	CA_CopyPlanes(map->planes, lwlib_X(pos), lwlib_Y(pos), yank->planes, 
		0, 0, lwlib_X(span) + 1, lwlib_Y(span) + 1);
}

void MapTool::pasteTile()
{
	MapTool_Map_t *map;

	map = &priv.maps[priv.mapIndex];

	switch (priv.yank.mode)
	{
	case MapTool_YankMode::Wall:
		CURMAPSPOT(0) = priv.yank.wall;
		CURMAPSPOT(1) = 0;
		CURMAPSPOT(2) = 0;
		break;
	case MapTool_YankMode::Actor:
		if (!CURMAPSPOT_ISSOLID())
		{
			CURMAPSPOT(1) = priv.yank.actor;
			if (priv.yank.wall == AMBUSHTILE)
			{
				CURMAPSPOT(0) = AMBUSHTILE;
			}
		}
		break;
	case MapTool_YankMode::FloorCeil:
		if (CURMAPSPOT_ISSOLID())
		{
			removeSolidWall();
		}

		CURMAPSPOT(2) = (CURMAPSPOT(2) & 0xff) + 
			(priv.yank.ceiling << 8);
		CURMAPSPOT(2) = (CURMAPSPOT(2) & 0xff00) + 
			priv.yank.floor;
		CURMAPSPOT(0) = priv.yank.wall;
		break;
	case MapTool_YankMode::WallPatch:
		if (CURMAPSPOT_ISSOLID())
		{
			CURMAPSPOT(1) = priv.yank.wallPatch;
		}
		else
		{
			CURMAPSPOT(0) = priv.yank.wall;
			CURMAPSPOT(1) = priv.yank.wallPatch;
			CURMAPSPOT(2) = 0;
		}
		break;
	case MapTool_YankMode::Picker:
		insertPicked();
		break;
	case MapTool_YankMode::TileBox:
		pasteTileBox();
		break;
	default:
		break;
	}
}

void MapTool::removeTile()
{
	MapTool_Map_t *map;

	map = &priv.maps[priv.mapIndex];

	if (CURMAPSPOT_ISSOLID() || CURMAPSPOT(0) == AMBUSHTILE)
	{
		removeSolidWall();
	}
	else
	{
		CURMAPSPOT(1) = 0;
	}
}

void MapTool::tileManip()
{
	MapTool_Map_t *map;

	if (!TILEINMAP(lwlib_X(priv.tilePos), lwlib_Y(priv.tilePos)))
	{
		return;
	}

	map = &priv.maps[priv.mapIndex];

	if (priv.commandSet.yankTile->pending())
	{
		yankTile();
	}
	if (priv.commandSet.pasteTile->pending())
	{
		pasteTile();
	}
	if (priv.commandSet.pushWall->pending())
	{
		CURMAPSPOT(1) = PUSHABLETILE;
	}
	if (priv.commandSet.ambush->pending())
	{
		CURMAPSPOT(0) = AMBUSHTILE;
	}
}

void MapTool::drawTileInfo()
{
	int tile;
	int type;
	word maptile;
	lwlib_String_t tmpStr;
	MapTool_Map_t *map;
	ED_Spawn_t *spawn;
	Point2i pos;

	if (priv.edit->isActive())
	{
		return;
	}

	map = &priv.maps[priv.mapIndex];

	pos = vec2i_zero();

	maptile = CURMAPSPOT(1);

	type = ED_ActorStatInfo(ed, maptile);
	if (type != -1)
	{
		lwlib_Snprintf(tmpStr, "statInfo.name: \"%s\"\n", 
			ED_STATINFO[type].name);
		pos = priv.font->printPosFit(pos, tmpStr);
	}

	spawn = ED_ActorSpawn(ed, maptile);
	if (spawn != NULL)
	{
		lwlib_Snprintf(tmpStr, "spawn.enemyName: \"%s\"\n", 
			spawn->enemyName);
		pos = priv.font->printPosFit(pos, tmpStr);
	}

	if (!CURMAPSPOT_ISSOLID())
	{
		maptile = CURMAPSPOT(0);
		if (maptile >= AREATILE && maptile < AREATILE + NUMAREAS)
		{
			lwlib_Snprintf(tmpStr, "areaCode: %d\n", 
				maptile - AREATILE);
			pos = priv.font->printPosFit(pos, tmpStr);
		}
	}

	if (lwlib_X(priv.tilePos) == 0 && lwlib_Y(priv.tilePos) == 0)
	{
		tile = MAPSPOT(0, 0, 0);
		lwlib_Snprintf(tmpStr, "fog color index: %d\n"
			"fog strength: %d\n", tile & 0xff, (tile & 0xff00) >> 8);
		pos = priv.font->printPosFit(pos, tmpStr);
	}
}

void MapTool::drawTileCursor()
{
	int i;
	Point2f pos;
	Point2f size;

	const int grid = priv.grid;

	for (i = 0; i < 2; i++)
	{
		lwlib_C(pos, i) = (lwlib_C(priv.tilePos, i) +
			lwlib_C(priv.tileMove, i) - 
			(priv.tileCursorSize / 2)) * grid;
		lwlib_C(size, i) = priv.tileCursorSize * grid;
	}

	priv.gl->drawBox((int)lwlib_X(pos),
		(int)lwlib_Y(pos), (int)lwlib_X(size), (int)lwlib_Y(size), 
		GAMEPAL_RGBA(LWLIB_MAPTOOL_BLUE, 0), BOX_USE_LINES);
}

void MapTool::drawTileBox()
{
	int x, y, w, h;
	Box2i box;
	Point2i pos;
	Vector2i span;
	lwlib_MapTool_Yank_t *yank;

	yank = &priv.yank;
	const int grid = priv.grid;

	if (yank->mode != MapTool_YankMode::TileBox)
	{
		return;
	}

	box = box_normalize(yank->tileBox);
	pos = lwlib_box_lower(box);
	span = box_span(box);

	if (yank->tileBoxForming)
	{
		x = lwlib_X(pos);
		y = lwlib_Y(pos);
	}
	else
	{
		x = lwlib_X(priv.tilePos);
		y = lwlib_Y(priv.tilePos);
	}
	w = lwlib_X(span) + 1;
	h = lwlib_Y(span) + 1;

#ifdef SDL_TRUE_COLOR
	priv.gl->drawBox(x * grid, 
		y * grid, w * grid, h * grid, GAMEPAL_RGBA(LWLIB_MAPTOOL_WHITE, 
		yank->tileBoxForming ? 128 : 192), false);
#endif
	priv.gl->drawBox(x * grid, 
		y * grid, w * grid, h * grid, GAMEPAL_RGBA(LWLIB_MAPTOOL_WHITE, 0),
		true);
}

void MapTool::drawTiles()
{
	int type;
	int tile;
	int lock;
	int dir;
	word maptile;
	bool vertical;
	int x, y;
	Img_t img;
	Gl *gl;
	Point2i pos;
	Vector2i span;
	const ED_Spawn_t *spawn;
	const ED_Enemy_t *enemy;
	lwlib_MapTool_Yank_t *yank;
	lwlib_MapTool_DrawCache_t *map;
	Point2i pan, p1, p2, glDims;

	map = &priv.drawCache;
	yank = &priv.yank;

	lwlib_Copy(map->planes, priv.maps[priv.mapIndex].planes);

	if (yank->mode == MapTool_YankMode::TileBox && 
		!yank->tileBoxForming)
	{
		pos = priv.tilePos;
		span = box_span(yank->tileBox);

		CA_CopyPlanes(map->planes, lwlib_X(pos), lwlib_Y(pos), 
			yank->planes, 0, 0, lwlib_X(span) + 1, lwlib_Y(span) + 1);
	}

	gl = priv.gl;

	const int grid = priv.grid;

	p1 = vec2i(0, 0);
	p2 = vec2i(64, 64);
	if (grid != 0)
	{
		pan = vec2f_int(vec2f(lwlib_ExpandVector2(
			priv.pan)));
		pan = vec_divide(pan, grid);
		pan = vec_neg(pan);

		glDims = priv.gl->getDims();

		p1 = pan;
		p2 = vec_add(p1, vec_divide(glDims, grid));
		p2 = vec_add(p2, vec2i(2, 2));

		p1 = vec_clip_minmax(p1, vec2i(0, 0), 
			vec2i(64, 64));
		p2 = vec_clip_minmax(p2, vec2i(0, 0),
			vec2i(64, 64));
	}

	for (y = lwlib_Y(p1); y < lwlib_Y(p2); y++)
	{
		for (x = lwlib_X(p1); x < lwlib_X(p2); x++)
		{
			maptile = MAPSPOT(x, y, 2);
			if (priv.ceilingMode)
			{
				tile = maptile >> 8;
			}
			else
			{
				tile = maptile & 0xff;
			}

			if (tile != 0 && !MAPSPOT_ISSOLID(x, y))
			{
				gl->setCurColor(vec4b(128, 128, 128, 255));

				img = priv.tileViewer->getTileImg(tile);
				if (img != NULL)
				{
					img->drawScaled(gl, x * grid, y * grid, grid, 
						grid, false);
				}

				gl->resetCurColor();
			}

			maptile = MAPSPOT(x, y, 0);
			if (maptile >= DOORTILE_FIRST && maptile <= DOORTILE_LAST)
			{
				vertical = (maptile % 2 == 0);
				lock = (maptile - DOORTILE_FIRST) / 2;
				tile = MAPSPOT(MAPSIZE - 1, MAPSIZE - 1, 0) +
					(((lock * 2) + dr_face_front) * 2) + (int)vertical;
				img = priv.tileViewer->getTileImg(tile);
				if (img != NULL)
				{
					img->drawScaled(gl, x * grid, y * grid, grid, 
						grid, false);
				}
			}
			else if
				(
					maptile > 0 && maptile < AMBUSHTILE &&
					!(
						MAPSPOT_ISSOLID(x - 1, y) &&
						MAPSPOT_ISSOLID(x + 1, y) &&
						MAPSPOT_ISSOLID(x, y - 1) &&
						MAPSPOT_ISSOLID(x, y + 1)
					)
				)
			{
				tile = horizwall[maptile];
				img = priv.tileViewer->getTileImg(tile);
				if (img != NULL)
				{
					img->drawScaled(gl, x * grid, y * grid, grid, 
						grid, false);
				}

				maptile = MAPSPOT(x, y, 1);
				if (ED_ActorIsWallPatch(ed, maptile))
				{
					tile = horizwall[ED_ActorToWallPatch(ed, maptile)];
					img = priv.tileViewer->getTileImg(tile);
					if (img != NULL)
					{
						img->drawScaled(gl, x * grid, y * grid, 
							grid, grid, true);
					}
				}
			}
			else if (maptile == AMBUSHTILE)
			{
				priv.gl->drawBox(x * grid, 
					y * grid, grid, grid, GAMEPAL_RGBA(LWLIB_MAPTOOL_ORANGE, 
					128), BOX_USE_LINES);
			}

			maptile = MAPSPOT(x, y, 1);

			type = ED_ActorStatInfo(ed, maptile);
			if (type != -1)
			{
				tile = ED_SPR(ED_STATINFO[type].picnum) +
					PMSpriteStart;
				img = priv.tileViewer->getTileImg(tile);
				if (img != NULL)
				{
					img->drawScaled(gl, x * grid, y * grid, grid, 
						grid, true);
				}
			}

			spawn = ED_ActorSpawn(ed, maptile);
			if (spawn != NULL)
			{
				tile = ED_SPR(ED_STATE(spawn->startState)->shapenum);
				if (tile != -1)
				{
					enemy = ED_Enemy(ed, spawn->enemyName);
					if (enemy != NULL)
					{
						tile += ED_SPR(enemy->spriteOffset.newStart) -
							ED_SPR(enemy->spriteOffset.oldStart);
						tile = ED_ReplaceSprite(enemy, tile);
					}

					tile += PMSpriteStart;
				}
				else
				{
					tile = PMSpriteStart;
				}

				img = priv.tileViewer->getTileImg(tile);
				if (img != NULL)
				{
					img->drawScaled(gl, x * grid, y * grid, grid, 
						grid, true);
				}

				if (spawn->dir >= 0 && spawn->dir < 4)
				{
					gl->drawBoxArrow(x * grid + (grid / 2), 
						y * grid + (grid / 2),
						grid / 4, grid / 8, grid / 4, 
						vec4b(0, 0xff, 0, 0xc0), spawn->dir);
				}
			}

			if (maptile >= ICONARROWS && 
				maptile < ICONARROWS + NUMICONARROWS)
			{
				gl->drawBoxArrow(x * grid + (grid / 2), 
					y * grid + (grid / 2),
					grid / 4, grid / 8, 0,
					vec4b(0xff, 0xff, 0xff, 0xc0), 
					(maptile - ICONARROWS) * 0.5);
			}

			if (maptile >= PLAYERSPAWNTILE && 
				maptile < PLAYERSPAWNTILE + 4)
			{
				priv.gl->drawBox(x * grid, 
					y * grid, grid, grid, GAMEPAL_RGBA(LWLIB_MAPTOOL_WHITE, 
					128), BOX_USE_LINES);

				dir = maptile - PLAYERSPAWNTILE;
				gl->drawBoxArrow(x * grid + (grid / 2), 
					y * grid + (grid / 2),
					grid / 4, grid / 8, grid / 4, 
					vec4b(0xff, 0xff, 0xff, 0xc0), 
					(1 - dir + 4) % 4
					);
			}
			else if (maptile == PUSHABLETILE)
			{
				priv.gl->drawBox(x * grid, 
					y * grid, grid, grid, GAMEPAL_RGBA(LWLIB_MAPTOOL_OTHER,
					128), BOX_USE_LINES);
			}
		}
	}

	if (priv.showGrid)
	{
		for (x = lwlib_X(p1); x < lwlib_X(p2) + 1; x++)
		{
			priv.gl->drawLine(x * grid, lwlib_Y(p1) * grid, 
				x * grid, lwlib_Y(p2) * grid, 
				GAMEPAL_RGBA(LWLIB_MAPTOOL_WHITE, 128));
		}
		for (y = lwlib_Y(p1); y < lwlib_Y(p2) + 1; y++)
		{
			priv.gl->drawLine(lwlib_X(p1) * grid, y * grid, 
				lwlib_X(p2) * grid, y * grid, 
				GAMEPAL_RGBA(LWLIB_MAPTOOL_WHITE, 128));
		}
	}
}

void MapTool::commandSet_AddToEdit(Edit *edit)
{
	lwlib_Edit_AddGroup(edit, "yankTile",
		priv.commandSet.yankTile->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "pasteTile",
		priv.commandSet.pasteTile->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "zoomIn",
		priv.commandSet.zoomIn->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "zoomOut",
		priv.commandSet.zoomOut->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "toggleCeilingMode",
		priv.commandSet.toggleCeilingMode->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "pushWall",
		priv.commandSet.pushWall->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "ambush",
		priv.commandSet.ambush->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "toggleGrid",
		priv.commandSet.toggleGrid->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "saveMaps",
		priv.commandSet.saveMaps->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "loadMaps",
		priv.commandSet.loadMaps->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "picker",
		priv.commandSet.picker->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "removeTile",
		priv.commandSet.removeTile->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "nextDir",
		priv.commandSet.nextDir->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "clearMap",
		priv.commandSet.clearMap->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "cutTile",
		priv.commandSet.cutTile->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "tileBox",
		priv.commandSet.tileBox->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "nextMap",
		priv.commandSet.nextMap->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "prevMap",
		priv.commandSet.prevMap->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "moveTileLeft",
		priv.commandSet.moveTileLeft->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "moveTileRight",
		priv.commandSet.moveTileRight->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "moveTileUp",
		priv.commandSet.moveTileUp->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "moveTileDown",
		priv.commandSet.moveTileDown->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "moveViewLeft",
		priv.commandSet.moveViewLeft->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "moveViewRight",
		priv.commandSet.moveViewRight->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "moveViewUp",
		priv.commandSet.moveViewUp->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "moveViewDown",
		priv.commandSet.moveViewDown->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "shiftTileLeft",
		priv.commandSet.shiftTileLeft->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "shiftTileRight",
		priv.commandSet.shiftTileRight->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "shiftTileUp",
		priv.commandSet.shiftTileUp->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "shiftTileDown",
		priv.commandSet.shiftTileDown->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "setFogColor",
		priv.commandSet.setFogColor->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "panView",
		priv.commandSet.panView->addToEdit(edit));
}

void MapTool::tileReplace_AddToEdit(Edit *edit)
{

	edit->addItem_int("sourceTileFirst",
		&priv.tileReplace.sourceTileFirst);
	edit->addItem_int("sourceTileLast",
		&priv.tileReplace.sourceTileLast);
	edit->addItem_int("destTile", &priv.tileReplace.destTile);
	edit->addItem_int("plane", &priv.tileReplace.plane);
	edit->addItem_bool("go", &priv.tileReplace.go);
}

void MapTool::shiftTile(int x, int y)
{
	yankTile();
	removeTile();
	lwlib_X(priv.tilePos) += x;
	lwlib_Y(priv.tilePos) += y;
	pasteTile();
}

void MapTool::checkTileReplace()
{
	int tile;
	int x, y;
	MapTool_Map_t *map;
	lwlib_MapTool_TileReplace_t *tileReplace;

	tileReplace = &priv.tileReplace;
	if (!tileReplace->go)
	{
		return;
	}

	tileReplace->go = false;

	map = &priv.maps[priv.mapIndex];

	for (y = 0; y < 64; y++)
	{
		for (x = 0; x < 64; x++)
		{
			if (tileReplace->plane != -1)
			{
				tile = MAPSPOT(x, y, tileReplace->plane);
				if (tile >= tileReplace->sourceTileFirst &&
					tile <= tileReplace->sourceTileLast)
				{
					tile = (tile - tileReplace->sourceTileFirst) + 
						tileReplace->destTile;
				}
				MAPSPOT(x, y, tileReplace->plane) = tile;
			}
			else
			{
				tile = MAPSPOT(x, y, 0);
				if (tile >= 64 && tile < 90)
				{
					MAPSPOT(x, y, 0) = 3;
					MAPSPOT(x, y, 1) = 471 + (tile - 64);
				}
			}
		}
	}
}

MapTool::MapTool(
	Gl *gl,
	TileViewer *tileViewer,
	Mouse_t mouse,
	lwlib_Keys_t keys,
	Edit *edit,
	Font *font,
	Picker *picker
	)
{
	lwlib_Zero(priv);

	priv.gl = gl;
	priv.tileViewer = tileViewer;
	priv.mouse = mouse;
	priv.keys = keys;
	priv.edit = edit;
	priv.font = font;
	priv.picker = picker;
	priv.loadMaps = true;

	priv.commandSet.yankTile = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.yankTile);
	priv.commandSet.pasteTile = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.pasteTile);
	priv.commandSet.zoomIn = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.zoomIn);
	priv.commandSet.zoomOut = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.zoomOut);
	priv.commandSet.toggleCeilingMode = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.toggleCeilingMode);
	priv.commandSet.pushWall = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.pushWall);
	priv.commandSet.ambush = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.ambush);
	priv.commandSet.toggleGrid = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.toggleGrid);
	priv.commandSet.saveMaps = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.saveMaps);
	priv.commandSet.loadMaps = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.loadMaps);
	priv.commandSet.picker = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.picker);
	priv.commandSet.removeTile = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.removeTile);
	priv.commandSet.nextDir = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.nextDir);
	priv.commandSet.clearMap = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.clearMap);
	priv.commandSet.cutTile = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.cutTile);
	priv.commandSet.tileBox = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.tileBox);
	priv.commandSet.nextMap = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.nextMap);
	priv.commandSet.prevMap = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.prevMap);
	priv.commandSet.moveTileLeft = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.moveTileLeft);
	priv.commandSet.moveTileRight = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.moveTileRight);
	priv.commandSet.moveTileUp = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.moveTileUp);
	priv.commandSet.moveTileDown = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.moveTileDown);
	priv.commandSet.moveViewLeft = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.moveViewLeft);
	priv.commandSet.moveViewRight = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.moveViewRight);
	priv.commandSet.moveViewUp = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.moveViewUp);
	priv.commandSet.moveViewDown = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.moveViewDown);
	priv.commandSet.shiftTileLeft = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.shiftTileLeft);
	priv.commandSet.shiftTileRight = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.shiftTileRight);
	priv.commandSet.shiftTileUp = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.shiftTileUp);
	priv.commandSet.shiftTileDown = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.shiftTileDown);
	priv.commandSet.setFogColor = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.setFogColor);
	priv.commandSet.panView = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.panView);
}

MapTool::~MapTool()
{
	vec_repeat(priv.commandSet.all, CommandPtr_t, delete x);
	vec_free(priv.commandSet.all, CommandPtr_t);
}

void MapTool::addToEdit(Edit *edit)
{
	edit->addItem_int("mapIndex", &priv.mapIndex);
	edit->addItem_int("tilePos.x", &lwlib_X(priv.tilePos));
	edit->addItem_int("tilePos.y", &lwlib_Y(priv.tilePos));
	edit->addItem_double("tileMoveSensitivity",
		&priv.tileMoveSensitivity);
	edit->addItem_int("grid", &priv.grid);
	edit->addItem_double("tileCursorSize", &priv.tileCursorSize);
	edit->addItem_bool("ceilingMode", &priv.ceilingMode);
	edit->addItem_int("planeToManip", &priv.planeToManip);
	edit->addItem_bool("fillArea", &priv.fillArea);
	edit->addItem_int("areaCode", &priv.areaCode);
	lwlib_Edit_AddGroup(edit, "commandSet",
		commandSet_AddToEdit(edit));
	lwlib_Edit_AddGroup(edit, "tileReplace",
		tileReplace_AddToEdit(edit));
	edit->addItem_bool("showGrid", &priv.showGrid);
}

void MapTool::updateEdit(Edit *edit)
{
	edit->addItem_group("map");
	map_UpdateEdit(edit);
	edit->navigateBack();

	edit->addItem_enum("fileFormat",
		(int *)&priv.fileFormat, MapTool_FileFormat_Str,
		NULL, lwlib_CountArray(MapTool_FileFormat_Str));
	edit->addItem_string("rawFileName", 
		priv.rawFileName, sizeof(priv.rawFileName));
}

void MapTool::updateFrame()
{
	int tile;
	ED_Spawn_t *spawn;
	MapTool_Map_t *map;
	lwlib_MapTool_Yank_t *yank;

	map = &priv.maps[priv.mapIndex];
	yank = &priv.yank;

	if (priv.loadMaps)
	{
		priv.loadMaps = false;
		loadMaps();
	}

	if (priv.saveMaps)
	{
		priv.saveMaps = false;
		saveMaps();
	}

	if (priv.fillArea)
	{
		priv.fillArea = false;
		lwlib_Zero(priv.areaTileVisit);
		setAreaTile(priv.tilePos);
	}

	if (!priv.edit->isActive())
	{
		vec_repeat(priv.commandSet.all, CommandPtr_t,
			x->processEvents(priv.keys, priv.mouse));
	}

	tilePosMotion();
	panMotion();
	tileManip();

	if (priv.commandSet.toggleCeilingMode->pending())
	{
		priv.ceilingMode = !priv.ceilingMode;
	}
	if (priv.commandSet.saveMaps->pending())
	{
		priv.saveMaps = true;
	}
	if (priv.commandSet.loadMaps->pending())
	{
		priv.loadMaps = true;
	}
	if (priv.commandSet.picker->pending())
	{
		priv.pickerPending = true;
		setYankMode(MapTool_YankMode::Picker);
	}

	if (priv.commandSet.removeTile->pending())
	{
		removeTile();
	}

	if (priv.commandSet.nextDir->pending())
	{
		if (!CURMAPSPOT_ISSOLID())
		{
			spawn = ED_ActorSpawn(ed, CURMAPSPOT(1));
			if (spawn != NULL && spawn->dir >= 0 && spawn->dir < 4)
			{
				if (spawn->dir == 3)
				{
					assert(spawn - &vec_at(ED_Spawns(ed), 0) >= 3);
					spawn -= 3;
					assert(spawn->dir == 0);
				}
				else
				{
					spawn++;
				}

				CURMAPSPOT(1) = spawn->actor;
			}

			tile = CURMAPSPOT(1);
			if (tile >= PLAYERSPAWNTILE && tile < PLAYERSPAWNTILE + 4)
			{
				CURMAPSPOT(1) = (((tile - PLAYERSPAWNTILE) + 1) % 4) + 
					PLAYERSPAWNTILE;
			}
			else if (tile >= ICONARROWS && 
				tile < ICONARROWS + NUMICONARROWS)
			{
				CURMAPSPOT(1) = (((tile - ICONARROWS) + 1) % 
					NUMICONARROWS) + ICONARROWS;
			}
		}
	}

	if (priv.commandSet.clearMap->pending())
	{
		lwlib_Zero(map->planes);
	}

	if (priv.commandSet.cutTile->pending())
	{
		yankTile();
		removeTile();
	}

	if (priv.commandSet.tileBox->pending())
	{
		setYankMode(MapTool_YankMode::TileBox);
		lwlib_box_lower(yank->tileBox) = 
			lwlib_box_upper(yank->tileBox) = priv.tilePos;
		yank->tileBoxForming = true;
	}

	if (yank->tileBoxForming)
	{
		lwlib_box_upper(yank->tileBox) = priv.tilePos;
	}

	if (priv.commandSet.nextMap->pending())
	{
		priv.mapIndex = (priv.mapIndex + 1) % NUMMAPS;
	}
	if (priv.commandSet.prevMap->pending())
	{
		priv.mapIndex = (priv.mapIndex + NUMMAPS - 1) % NUMMAPS;
	}

	if (priv.commandSet.moveTileLeft->pending())
	{
		lwlib_X(priv.tilePos)--;
	}
	if (priv.commandSet.moveTileRight->pending())
	{
		lwlib_X(priv.tilePos)++;
	}
	if (priv.commandSet.moveTileUp->pending())
	{
		lwlib_Y(priv.tilePos)--;
	}
	if (priv.commandSet.moveTileDown->pending())
	{
		lwlib_Y(priv.tilePos)++;
	}

	if (priv.commandSet.moveViewLeft->pending())
	{
		lwlib_X(priv.pan) += priv.grid;
		lwlib_X(priv.tilePos)--;
	}
	if (priv.commandSet.moveViewRight->pending())
	{
		lwlib_X(priv.pan) -= priv.grid;
		lwlib_X(priv.tilePos)++;
	}
	if (priv.commandSet.moveViewUp->pending())
	{
		lwlib_Y(priv.pan) += priv.grid;
		lwlib_Y(priv.tilePos)--;
	}
	if (priv.commandSet.moveViewDown->pending())
	{
		lwlib_Y(priv.pan) -= priv.grid;
		lwlib_Y(priv.tilePos)++;
	}

	if (priv.commandSet.shiftTileLeft->pending())
	{
		shiftTile(-1, 0);
	}
	if (priv.commandSet.shiftTileRight->pending())
	{
		shiftTile(1, 0);
	}
	if (priv.commandSet.shiftTileUp->pending())
	{
		shiftTile(0, -1);
	}
	if (priv.commandSet.shiftTileDown->pending())
	{
		shiftTile(0, 1);
	}

	if (priv.commandSet.setFogColor->pending())
	{
		tile = priv.picker->tileSelector();
		MAPSPOT(0, 0, 0) = (MAPSPOT(0, 0, 0) & 0xff00) | (tile & 0xff);
	}

	if (priv.commandSet.toggleGrid->pending())
	{
		priv.showGrid = !priv.showGrid;
	}

	checkTileReplace();
}

void MapTool::draw(Gl *gl)
{
	gl->pushMatrix();

	gl->translateMatrix(lwlib_ExpandVector(priv.pan));

	const int grid = priv.grid;

	drawTiles();

	priv.gl->drawBox(0, 0, 64 * grid, 64 * grid, 
		GAMEPAL_RGBA(LWLIB_MAPTOOL_WHITE, 0), true);

	priv.gl->drawBox(lwlib_X(priv.tilePos) * grid, 
		lwlib_Y(priv.tilePos) * grid, grid, grid, GAMEPAL_RGBA(
		LWLIB_MAPTOOL_RED, 192), BOX_USE_LINES);

	drawTileCursor();
	drawTileBox();

	gl->popMatrix();

	drawTileInfo();
}

bool MapTool::pickerPending()
{
	bool pending;

	pending = priv.pickerPending;
	if (pending)
	{
		priv.pickerPending = false;
	}

	return pending;
}

int MapTool::getPickerDoorTile(int lock)
{
	MapTool_Map_t *map;

	map = &priv.maps[priv.mapIndex];

	if (lock >= (DOORTILE_LAST - DOORTILE_FIRST + 1) / 2)
	{
		return -1;
	}

	return MAPSPOT(MAPSIZE - 1, MAPSIZE - 1, 0) + 
		(((lock * 2) + dr_face_front) * 2) + dr_orient_horz;
}
