/*************************************************************************
** Wolf3D Legacy
** Copyright (C) 2012 by LinuxWolf - Team RayCAST
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**************************************************************************
** LinuxWolf Library for Wolf3D Legacy
*************************************************************************/

#ifndef LWLIB_MOUSE_H
#define LWLIB_MOUSE_H

#include <SDL/SDL_events.h>
#include "lw_vec.h"
#include "lw_edit.h"

namespace lwlib
{
	namespace Mouse_Button
	{
		enum e
		{
			Left,
			Middle,
			Right,
			Max,
		};
	}

	lwlib_EnumStrExtern(Mouse_Button);

	namespace Mouse_Event
	{
		enum e
		{
			Down,
			Pressed,
			Released,
			Max,
		};
	}

	lwlib_EnumStrExtern(Mouse_Event);

	namespace Mouse_ActiveFlag
	{
		enum e
		{
			AppFocus,
			EditMode,
			Max,
		};
	}

	class Mouse
	{
		struct
		{
			bool ignore;
			bool active;
			Point2i vidDims;
			int stuck;
			Point2i oldMPos;
			bool filter;
			double sensitivity;
			Point2i move;
			bool debugMove;
			Edit *edit;
			int buttonState;
			int newButtons;
			int activeStatus;
		} priv;

	public:
		explicit Mouse(Point2i vidDims);

		~Mouse();

		void setIgnore(bool ignore);

		void activate();

		void deactivate();

		void pollState();

		void addToEdit(Edit *edit);

		Point2i getMove();

		void updateFrame();

		void setEdit(Edit *edit);

		bool checkButtonStatus(Mouse_Button::e button, 
			Mouse_Event::e event);

		void setActiveFlagState(Mouse_ActiveFlag::e activeFlag, 
			bool state);
	};

#ifndef LWLIB_MOUSE_PTRTYPE
#define LWLIB_MOUSE_PTRTYPE
	typedef Mouse *Mouse_t;
#endif
}

#endif
