/*************************************************************************
** Wolf3D Legacy
** Copyright (C) 2012 by LinuxWolf - Team RayCAST
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**************************************************************************
** LinuxWolf Library for Wolf3D Legacy
*************************************************************************/

#ifndef LWLIB_OCTREE_H
#define LWLIB_OCTREE_H

#include "lw_vec.h"
#include "lw_misc.h"
#include "lw_render.h"
#include "lw_csg.h"

#define lwlib_Octree_Valid(octree) ((octree) != NULL)

namespace lwlib
{
	namespace Octree_RenderAlgo
	{
		enum e
		{
			SolidBox,
			Surface,
			Cells,
		};
	}

	namespace Octree_SetOp
	{
		enum e
		{
			Union,
			Intersection,
		};
	}

	typedef struct Octree_Node_s
	{
		int size;
		Box3i cellBox;
		Point3i cellBoxCenter;
		bool isParent;
		bool cellState;
		struct Octree_Node_s *childNodes[8];
		struct Octree_Node_s *parentNode;
		struct Octree_Node_s *sideNodes[6];
		int childNodeIndex;
	} Octree_Node_t;

	class Octree
	{
		struct
		{
			Octree_Node_t *topNode;
			int sliceAxis;
			int slicePos;
			bool sliceEnabled;
			bool cullingEnabled;
			Point3i cullingViewPos;
		} priv;

		void renderSolidBox(lwlib_Render_t render);

		void renderSurface(lwlib_Render_t render);

		void renderCells(lwlib_Render_t render);

		explicit Octree(Octree_Node_t *topNode);
		
	public:
		explicit Octree(int size);

		~Octree();

		Octree *copy();

		void expand(int size);

		void setState(Box3i cellBox, bool cellState);

		void setCellState(Point3i cellPos, bool cellState);

		void setCsgState(Csg *csg, bool cellState);

		bool queryState(Point3i cellPos);

		void applyBrush(Octree *brushOctree, bool cellState);

		void applyNegate();

		void applySetOp(Octree *octree, Octree_SetOp::e setOp);

		void scale(int level);

		void render(Octree_RenderAlgo::e algo, lwlib_Render_t render);

		void setSlice(bool sliceEnabled, int sliceAxis, int slicePos);

		int countNodes();

		void debugPrint();

		int size();

		void setCullingEnabled(bool cullingEnabled);

		void setCullingViewPos(Point3i cullingViewPos);
	};
}

#endif


