/************************************************************************* ** Wolf3D Legacy
** Copyright (C) 2012 by LinuxWolf - Team RayCAST
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**************************************************************************
** LinuxWolf Library for Wolf3D Legacy
*************************************************************************/

#include "lw_picker.h"
#include "lw_img.h"
#include "lw_intmap.h"
#include "lw_tileviewer.h"
#include "lw_command.h"
#include "wl_def.h"
using namespace lwlib;

#define LWLIB_PICKER_BORDER 0.015
#define LWLIB_PICKER_MAXACROSS 20
#define LWLIB_PICKER_MAXDOWN 15
#define LWLIB_PICKER_RED 4
#define LWLIB_PICKER_BLUE 7

#define deln lwlib_CallDelegateNoArgs
#define del lwlib_CallDelegate

typedef lwlib_Byte_t byte;

lwlib_EnumStr(
	Picker_ObjType,
	"Wall",
	"Floor",
	"Ceiling",
	"WallPatch",
	"StatActor",
	"EnemyActor",
	"Door",
	"Area",
	"ColorIndex",
);

Img_t Picker::getTileImg(int tile)
{
	byte *source;
	Img_t img;
	lwlib_DeclDelegateAPI(Picker_DelegateAPI_t, this);

	lwlib_Zero(img);
	if (tile < 0)
	{
		return img;
	}

	switch (priv.curObjType)
	{
	case Picker_ObjType::Wall:
		if (tile >= DOORTILE_FIRST)
		{
			tile = -1;
		}
		else
		{
			tile = horizwall[tile];
		}
		break;
	case Picker_ObjType::Floor:
	case Picker_ObjType::Ceiling:
		if (tile >= PMSpriteStart || 
			tile >= ED_GetVSWAPInfo(ed)->skyStart)
		{
			tile = -1;
		}
		break;
	case Picker_ObjType::WallPatch:
		if (tile >= ED_GetTileInfo(ed)->lastWallPatchActor - 
			ED_GetTileInfo(ed)->firstWallPatchActor + 1)
		{
			tile = -1;
		}
		else
		{
			tile = horizwall[ED_FirstPatchTile(ed) + tile];
			if (tile >= PMSpriteStart || 
				tile >= ED_GetVSWAPInfo(ed)->skyStart)
			{
				tile = -1;
			}
		}
		break;
	case Picker_ObjType::StatActor:
		tile = ED_StatInfoPickerTile(ed, tile);
		break;
	case Picker_ObjType::EnemyActor:
		tile = ED_SpawnPickerTile(ed, tile);
		break;
	case Picker_ObjType::Door:
		tile = del(getDoorTile, tile);
		break;
	case Picker_ObjType::Area:
		if (tile >= NUMAREAS)
		{
			tile = -1;
		}
		else
		{
			tile = PMSpriteStart;
		}
		break;
	case Picker_ObjType::ColorIndex:
		if (tile < 256)
		{
			if (priv.colorIndexImgs[tile] == NULL)
			{
				source = lwlib_CallocMany(byte, 
					TEXTURESIZE * TEXTURESIZE);
				memset(source, tile, TEXTURESIZE * TEXTURESIZE);
				img = new Img();
				img->fromSource(source, TEXTURESIZE, 
					TEXTURESIZE, 1, TEXTURESIZE, true, priv.palette);
#ifdef SDL_TRUE_COLOR
				img->expandPalette(true);
#endif
				priv.colorIndexImgs[tile] = img;
			}
			img = priv.colorIndexImgs[tile];
		}
		tile = -1;
		break;
	}

	if (tile != -1)
	{
		img = priv.tileViewer->getTileImg(tile);
	}

	return img;
}

void Picker::drawTile(int tile, int cx, int cy, int sw, int sh)
{
	Gl *gl;
	Img_t img;

	gl = priv.gl;

	img = getTileImg(tile);
	if (img != NULL)
	{
		img->drawScaled(gl, cx, cy, sw, sh, false);
	}
}

void Picker::drawTileInfo()
{
	int tile;
	lwlib_String_t infoStr;
	lwlib_String_t tmpStr[3];
	ED_StatInfo_t *statInfo;

	if (priv.edit->isActive())
	{
		return;
	}

	tile = priv.tileSelector;

	lwlib_Zero(infoStr);

	switch (priv.curObjType)
	{
	case Picker_ObjType::Wall:
		lwlib_Snprintf(infoStr, "wall: %d", tile);
		break;
	case Picker_ObjType::Floor:
		lwlib_Snprintf(infoStr, "floor: %d", tile);
		break;
	case Picker_ObjType::Ceiling:
		lwlib_Snprintf(infoStr, "ceiling: %d", tile);
		break;
	case Picker_ObjType::WallPatch:
		lwlib_Snprintf(infoStr, "wall patch: %d", tile);
		break;
	case Picker_ObjType::StatActor:
		lwlib_Snprintf(infoStr, "static actor: %d\n", tile);

		statInfo = &ED_STATINFO[tile];
		lwlib_Copy(tmpStr[0], infoStr);
		lwlib_Snprintf(tmpStr[1], "name = \"%s\"\n", statInfo->name);
		lwlib_Snprintf(infoStr, "%s%s", tmpStr[0], tmpStr[1]);

		if (statInfo->bonus[0] != '\0')
		{
			lwlib_Copy(tmpStr[0], infoStr);
			lwlib_Snprintf(tmpStr[1], "bonus = \"%s\"\n", 
				statInfo->bonus);
			lwlib_Snprintf(infoStr, "%s%s", tmpStr[0], tmpStr[1]);
		}
		break;
	case Picker_ObjType::EnemyActor:
		lwlib_Snprintf(infoStr, "%s", 
			ED_SpawnPickerInfoString(ed, tile).v);
		break;
	case Picker_ObjType::Door:
		lwlib_Snprintf(infoStr, "door: %d", tile);
		break;
	case Picker_ObjType::Area:
		lwlib_Snprintf(infoStr, "area: %d\n", tile);

		if (tile < surpriseatk_max)
		{
			lwlib_Copy(tmpStr[0], infoStr);
			lwlib_Snprintf(tmpStr[1], "surpriseatk = \"%s\"\n", 
				surpriseatktype_Str[tile]);
			lwlib_Snprintf(infoStr, "%s%s", tmpStr[0], tmpStr[1]);
		}

		if (tile < villager_max)
		{
			lwlib_Copy(tmpStr[0], infoStr);
			lwlib_Snprintf(tmpStr[1], "villager = \"%s\"\n", 
				villagertype_Str[tile]);
			lwlib_Snprintf(infoStr, "%s%s", tmpStr[0], tmpStr[1]);
		}
		break;
	case Picker_ObjType::ColorIndex:
		lwlib_Snprintf(infoStr, "color index: %d (0x%02X)\n", tile, tile);

		lwlib_Copy(tmpStr[0], infoStr);
		lwlib_Snprintf(tmpStr[1], "(r,g,b) = (%d,%d,%d)\n", 
			gamepal[tile].r, gamepal[tile].g, gamepal[tile].b);
		lwlib_Snprintf(infoStr, "%s%s", tmpStr[0], tmpStr[1]);
		break;
	}

	if (infoStr[0] != '\0')
	{
		priv.font->printPosFit(vec2i(0, 0), infoStr);
	}
}

void Picker::handleObjMove()
{

	if (priv.objMove == 0)
	{
		return;
	}

	priv.tileSelector = ED_MoveStatInfo(ed, priv.tileSelector, 
		priv.objMove);
	priv.objMove = 0;
}

static void lwlib_Picker_CommandSet_AddToEdit(
	Picker_CommandSet_t *commandSet, Edit *edit)
{
	lwlib_Edit_AddGroup(edit, "wall",
		commandSet->wall->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "floor",
		commandSet->floor->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "ceiling",
		commandSet->ceiling->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "wallPatch",
		commandSet->wallPatch->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "statActor",
		commandSet->statActor->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "enemyActor",
		commandSet->enemyActor->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "door",
		commandSet->door->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "area",
		commandSet->area->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "colorIndex",
		commandSet->colorIndex->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "nextPage",
		commandSet->nextPage->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "prevPage",
		commandSet->prevPage->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "mapToolPending",
		commandSet->mapToolPending->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "moveLeft",
		commandSet->moveLeft->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "moveRight",
		commandSet->moveRight->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "moveDown",
		commandSet->moveDown->addToEdit(edit));
	lwlib_Edit_AddGroup(edit, "moveUp",
		commandSet->moveUp->addToEdit(edit));
}

Picker::Picker(
	lwlib_Keys_t keys,
	Gl *gl,
	Font *font,
	Edit *edit,
	TileViewer *tileViewer,
	Mouse_t mouse,
	SDL_Color *palette,
	Picker_DelegateAPI_t delegateAPI
	)
{
	lwlib_Zero(priv);

	priv.keys = keys;
	priv.gl = gl;
	priv.font = font;
	priv.edit = edit;
	priv.tileViewer = tileViewer;
	priv.mouse = mouse;
	priv.palette = palette;
	priv.delegateAPI = delegateAPI;
	priv.tilesAcross = 4;
	priv.tilesDown = 1;
	priv.tileMoveSensitivity = 0.005;
	priv.tileCursorSize = 0.2;
	priv.commandSet.wall = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.wall);
	priv.commandSet.floor = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.floor);
	priv.commandSet.ceiling = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.ceiling);
	priv.commandSet.wallPatch = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.wallPatch);
	priv.commandSet.statActor = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.statActor);
	priv.commandSet.enemyActor = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.enemyActor);
	priv.commandSet.door = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.door);
	priv.commandSet.area = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.area);
	priv.commandSet.colorIndex = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.colorIndex);
	priv.commandSet.nextPage = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.nextPage);
	priv.commandSet.prevPage = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.prevPage);
	priv.commandSet.mapToolPending = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.mapToolPending);
	priv.commandSet.moveLeft = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.moveLeft);
	priv.commandSet.moveRight = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.moveRight);
	priv.commandSet.moveDown = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.moveDown);
	priv.commandSet.moveUp = new Command();
	vec_push(priv.commandSet.all, CommandPtr_t,
		priv.commandSet.moveUp);
}

Picker::~Picker()
{
	int i;
	for (i = 0; i < 256; i++)
	{
		delete priv.colorIndexImgs[i];
	}
	vec_repeat(priv.commandSet.all, CommandPtr_t, delete x);
	vec_free(priv.commandSet.all, CommandPtr_t);
}

void Picker::addToEdit(Edit *edit)
{

	edit->addItem_int("tileSelector", &priv.tileSelector);
	edit->addItem_int("tilesAcross", &priv.tilesAcross);
	edit->addItem_int("tilesDown", &priv.tilesDown);
	edit->addItem_enum("objType",
		(int *)&priv.objType, Picker_ObjType_Str,
		NULL, Picker_ObjType_StrCount);
	lwlib_Edit_AddGroup(edit, "commandSet", 
		lwlib_Picker_CommandSet_AddToEdit(&priv.commandSet, edit));
}

void Picker::updateEdit(Edit *edit)
{
	ED_WallAttr_t *wallAttr;

	edit->addItem_int("objMove", &priv.objMove);
	edit->setItemEnabled("objMove", priv.curObjType == 
		Picker_ObjType::StatActor);

	if (priv.curObjType == Picker_ObjType::Wall)
	{
		wallAttr = ED_WallAttr(ed, priv.tileSelector);
		if (wallAttr)
		{
			edit->addItem_string("ricochetName", 
				wallAttr->ricochetName, sizeof(wallAttr->ricochetName));
		}
	}
}

void Picker::updateFrame()
{
	int i;
	int goRight, goLeft, goUp, goDown;
	int tileSelector;
	int nw, nh;
	int tilesPerPage;
	int page, cell;
	Point2i mouseMove;
	Img_t img;

	handleObjMove();

	if (priv.curObjType != priv.objType)
	{
		priv.curObjType = priv.objType;
		priv.tileSelector = 0;
	}

	if (priv.commandSet.wall->pending())
	{
		priv.objType = Picker_ObjType::Wall;
	}
	if (priv.commandSet.floor->pending())
	{
		priv.objType = Picker_ObjType::Floor;
	}
	if (priv.commandSet.ceiling->pending())
	{
		priv.objType = Picker_ObjType::Ceiling;
	}
	if (priv.commandSet.wallPatch->pending())
	{
		priv.objType = Picker_ObjType::WallPatch;
	}
	if (priv.commandSet.statActor->pending())
	{
		priv.objType = Picker_ObjType::StatActor;
	}
	if (priv.commandSet.enemyActor->pending())
	{
		priv.objType = Picker_ObjType::EnemyActor;
	}
	if (priv.commandSet.door->pending())
	{
		priv.objType = Picker_ObjType::Door;
	}
	if (priv.commandSet.area->pending())
	{
		priv.objType = Picker_ObjType::Area;
	}
	if (priv.commandSet.colorIndex->pending())
	{
		priv.objType = Picker_ObjType::ColorIndex;
	}

	if (!priv.edit->isActive())
	{
		vec_repeat(priv.commandSet.all, CommandPtr_t,
			x->processEvents(priv.keys, priv.mouse));

		mouseMove = priv.mouse->getMove();
	}
	else
	{
		mouseMove = vec2i_zero();
	}

	for (i = 0; i < 2; i++)
	{
		if (lwlib_C(mouseMove, i) == 0)
		{
			continue;
		}

		lwlib_C(priv.tileMove, i) += (double)lwlib_C(mouseMove, i) *
			priv.tileMoveSensitivity;
		while (lwlib_C(priv.tileMove, i) >= 1.0)
		{
			lwlib_C(priv.tilePos, i)++;
			lwlib_C(priv.tileMove, i)--;
		}
		while (lwlib_C(priv.tileMove, i) < 0.0)
		{
			lwlib_C(priv.tilePos, i)--;
			lwlib_C(priv.tileMove, i)++;
		}
	}

	tileSelector = priv.tileSelector;
	nw = lwlib_CLIP(priv.tilesAcross, 1, LWLIB_PICKER_MAXACROSS - 1);
	nh = lwlib_CLIP(priv.tilesDown, 1, LWLIB_PICKER_MAXDOWN - 1);
	tilesPerPage = nw * nh;
	page = tileSelector / tilesPerPage;
	cell = tileSelector % tilesPerPage;

	goRight = goLeft = goUp = goDown = 0;

	if (lwlib_X(priv.tilePos) > 0)
	{
		goRight += lwlib_X(priv.tilePos);
	}
	if (lwlib_X(priv.tilePos) < 0)
	{
		goLeft -= lwlib_X(priv.tilePos);
	}
	if (lwlib_Y(priv.tilePos) > 0)
	{
		goDown += lwlib_Y(priv.tilePos);
	}
	if (lwlib_Y(priv.tilePos) < 0)
	{
		goUp -= lwlib_Y(priv.tilePos);
	}

	priv.tilePos = vec2i_zero();

	if (priv.commandSet.moveLeft->pending())
	{
		goLeft++;
	}
	if (priv.commandSet.moveRight->pending())
	{
		goRight++;
	}
	if (priv.commandSet.moveDown->pending())
	{
		goDown++;
	}
	if (priv.commandSet.moveUp->pending())
	{
		goUp++;
	}

	while (goRight)
	{
		if ((cell % nw) == nw - 1)
		{
			tileSelector = (page + 1) * tilesPerPage + ((cell / nw) * nw);
		}
		else
		{
			tileSelector++;
		}
		goRight--;
	}

	while (goLeft)
	{
		if ((cell % nw) == 0)
		{
			tileSelector = (page - 1) * tilesPerPage + ((cell / nw) * nw + nw - 1);
		}
		else
		{
			tileSelector--;
		}
		goLeft--;
	}

	while (goDown)
	{
		tileSelector += nw;
		goDown--;
	}

	while (goUp)
	{
		tileSelector -= nw;
		goUp--;
	}

	if (priv.commandSet.nextPage->pending())
	{
		tileSelector += tilesPerPage;
	}

	if (priv.commandSet.prevPage->pending())
	{
		tileSelector -= tilesPerPage;
	}

	if (priv.commandSet.mapToolPending->pending())
	{
		priv.mapToolPending = true;
	}

	img = getTileImg(tileSelector);
	if (img != NULL)
	{
		priv.tileSelector = tileSelector;
	}
}

void Picker::draw(Gl *gl)
{
	int x, y;
	int sw, sh, cx, cy;
	int nw, nh;
	int bw, bh;
	int px, py;
	int tileStart, tileRem;

	nw = lwlib_CLIP(priv.tilesAcross, 1, LWLIB_PICKER_MAXACROSS - 1);
	nh = lwlib_CLIP(priv.tilesDown, 1, LWLIB_PICKER_MAXDOWN - 1);

	sw = priv.gl->getWidth();
	sh = priv.gl->getHeight();

	bw = (int)(LWLIB_PICKER_BORDER * (double)sw);
	bh = (int)(LWLIB_PICKER_BORDER * (double)sh);

	cx = (sw / 2);
	cy = (sh / 2);

	sw = TEXTURESIZE;
	cx = cx - (sw + 2 * bw) * nw / 2;

	sh = TEXTURESIZE;
	cy = cy - (sh + 2 * bh) * nh / 2;

	tileStart = (priv.tileSelector / (nw * nh)) * (nw * nh);
	tileRem = priv.tileSelector % (nw * nh);

	for (x = 0; x < nw; x++)
	{
		for (y = 0; y < nh; y++)
		{
			px = cx + (sw + 2 * bw) * x + bw;
			py = cy + (sh + 2 * bh) * y + bh;

			if (y * nw + x == tileRem)
			{
				priv.gl->drawBox(px - bw, 
					py - bh, sw + 2 * bw, 
					sh + 2 * bh, GAMEPAL_RGBA(LWLIB_PICKER_RED, 0), false);
			}

			drawTile(tileStart + y * nw + x, px, py, sw, sh);

			if (y * nw + x == tileRem)
			{
				priv.gl->drawBox(px - bw + (lwlib_X(priv.tileMove) * sw),
					py - bh + (lwlib_Y(priv.tileMove) * sh), 
					priv.tileCursorSize * sw, 
					priv.tileCursorSize * sh, 
					GAMEPAL_RGBA(LWLIB_PICKER_BLUE, 0), false);
			}
		}
	}

	drawTileInfo();

	if (!priv.edit->isActive())
	{
	}
}

bool Picker::mapToolPending()
{
	bool pending;

	pending = priv.mapToolPending;
	if (pending)
	{
		priv.mapToolPending = false;
	}

	return pending;
}

int Picker::tileSelector()
{
	return priv.tileSelector;
}

Picker_ObjType::e Picker::getObjType()
{
	return priv.curObjType;
}
