/*************************************************************************
** Wolf3D Legacy
** Copyright (C) 2012 by LinuxWolf - Team RayCAST
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**************************************************************************
** LinuxWolf Library for Wolf3D Legacy
*************************************************************************/

#ifndef LWLIB_PICKER_H
#define LWLIB_PICKER_H

#include "lw_img.h"
#include "lw_font.h"
#include "lw_keys.h"
#include "lw_edit.h"
#include "lw_tool.h"
#include "lw_tileviewer.h"
#include "lw_mouse.h"
#include "lw_misc.h"
#include "lw_gl.h"
#include "lw_command.h"

namespace lwlib
{
	namespace Picker_ObjType
	{
		enum e
		{
			Wall,
			Floor,
			Ceiling,
			WallPatch,
			StatActor,
			EnemyActor,
			Door,
			Area,
			ColorIndex,
		};
	}

	lwlib_DefineDelegate(Picker_DelegateGetDoorTile_t,
		// ret
		int,
		// params
		void *obj, int lock
		);

	typedef struct Picker_DelegateAPI_s
	{
		Picker_DelegateGetDoorTile_t getDoorTile;
	} Picker_DelegateAPI_t;

	typedef struct Picker_CommandSet_s
	{
		Command *wall;
		Command *floor;
		Command *ceiling;
		Command *wallPatch;
		Command *statActor;
		Command *enemyActor;
		Command *door;
		Command *area;
		Command *colorIndex;
		Command *nextPage;
		Command *prevPage;
		Command *mapToolPending;
		Command *moveLeft;
		Command *moveRight;
		Command *moveDown;
		Command *moveUp;
		CommandPtrVec_t all;
	} Picker_CommandSet_t;

	class Picker : public Tool
	{
		struct
		{
			lwlib_Keys_t keys;
			Gl *gl;
			Font *font;
			Edit *edit;
			TileViewer *tileViewer;
			Mouse_t mouse;
			SDL_Color *palette;
			Picker_DelegateAPI_t delegateAPI;
			int tileSelector;
			int tilesAcross;
			int tilesDown;
			Picker_ObjType::e objType;
			Picker_ObjType::e curObjType;
			int objMove;
			Point2i tilePos;
			Point2f tileMove;
			double tileMoveSensitivity;
			double tileCursorSize;
			Picker_CommandSet_t commandSet;
			bool mapToolPending;
			Img_t colorIndexImgs[256];
		} priv;

		Img_t getTileImg(int tile);

		void drawTile(int tile, int cx, int cy, int sw, int sh);

		void drawTileInfo();

		void handleObjMove();

	public:
		Picker(
			lwlib_Keys_t keys,
			Gl *gl,
			Font *font,
			Edit *edit,
			TileViewer *tileViewer,
			Mouse_t mouse,
			SDL_Color *palette,
			Picker_DelegateAPI_t delegateAPI
			);

		virtual ~Picker();

		virtual void addToEdit(Edit *edit);

		virtual void updateEdit(Edit *edit);

		virtual void updateFrame();

		virtual void draw(Gl *gl);

		bool mapToolPending();

		int tileSelector();

		Picker_ObjType::e getObjType();
	};
}

#endif
