/*************************************************************************
** Wolf3D Legacy
** Copyright (C) 2012 by LinuxWolf - Team RayCAST
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**************************************************************************
** LinuxWolf Library for Wolf3D Legacy
*************************************************************************/

#include "lw_picviewer.h"
#include "lw_img.h"
#include "lw_intmap.h"
#include "wl_def.h"
using namespace lwlib;

#define lwlib_PicViewer_KeyAction(name) lwlib_PicViewer_KeyAction_ ## name
#define lwlib_PicViewer_KeyAction_NextPic "l"
#define lwlib_PicViewer_KeyAction_PrevPic "h"

#define lwlib_PicViewer_KeyPressed(name) lwlib_Keys_CheckEvent(priv.keys, \
	lwlib_PicViewer_KeyAction(name), lwlib_Keys_Event(Pressed))
#define lwlib_PicViewer_KeyReleased(name) lwlib_Keys_CheckEvent(priv.keys, \
	lwlib_PicViewer_KeyAction(name), lwlib_Keys_Event(Released))
#define lwlib_PicViewer_KeyRepeat(name) lwlib_Keys_CheckEvent(priv.keys, \
	lwlib_PicViewer_KeyAction(name), lwlib_Keys_Event(Repeat))
#define lwlib_PicViewer_KeyFastRepeat(name) lwlib_Keys_CheckEvent(priv.keys, \
	lwlib_PicViewer_KeyAction(name), lwlib_Keys_Event(FastRepeat))
#define lwlib_PicViewer_KeyDown(name) lwlib_Keys_CheckEvent(priv.keys, \
	lwlib_PicViewer_KeyAction(name), lwlib_Keys_Event(Down))

typedef lwlib_Byte_t byte;

static void lwlib_PicViewer_MemToLatch(byte *source, 
	int width, int height, byte *dest)
{
	int pitch = width;
	for(int ysrc = 0; ysrc < height; ysrc++)
	{
		for(int xsrc = 0; xsrc < width; xsrc++)
		{
			dest[ysrc * pitch + xsrc] = 
				source[(ysrc * (width >> 2) + (xsrc >> 2))
					+ (xsrc & 3) * (width >> 2) * height];
		}
	}
}

void PicViewer::drawPic(int pic, int cx, int cy, int sw, int sh)
{
	Gl *gl;
	Img_t img;

	gl = priv.gl;

	img = priv.picImg;
	if (img != NULL)
	{
		img->drawScaled(gl, cx, cy, sw, sh, false);
	}
}

PicViewer::PicViewer()
{
	lwlib_Zero(priv);
}

PicViewer::~PicViewer()
{
	delete priv.picImg;
}

void PicViewer::addToEdit(Edit *edit)
{

	edit->addItem_int("picSelector", &priv.picSelector);
}

void PicViewer::updateEdit(Edit *edit)
{
}

void PicViewer::updateFrame()
{

	if (priv.picSelector != priv.oldPicSelector)
	{
		priv.oldPicSelector = priv.picSelector;

		delete priv.picImg;
		priv.picImg = getPicImg(priv.picSelector);
	}

	if (priv.edit->isActive())
	{
		return;
	}

	if (lwlib_PicViewer_KeyRepeat(NextPic))
	{
		priv.picSelector++;
	}

	if (lwlib_PicViewer_KeyRepeat(PrevPic))
	{
		priv.picSelector--;
	}
}

void PicViewer::draw(Gl *gl)
{
	int sw, sh, cx, cy;

	if (priv.picImg == NULL)
	{
		return;
	}

	sw = priv.gl->getWidth();
	sh = priv.gl->getHeight();

	cx = (sw / 2);
	cy = (sh / 2);

	sw = priv.picImg->getWidth();
	cx = cx - (sw / 2);

	sh = priv.picImg->getHeight();
	cy = cy - (sh / 2);

	drawPic(priv.picSelector, cx, cy, sw, sh);
}

Img_t PicViewer::getPicImg(int pic)
{
	int width, height;
    byte *source;
	Img_t img;

	lwlib_Zero(img);

	if (pic - STARTPICS < 0 || pic - STARTPICS >= NUMPICS)
	{
		return img;
	}

	width = pictable[pic - STARTPICS].width;
	height = pictable[pic - STARTPICS].height;

	CA_CacheGrChunk(pic);

	source = lwlib_CallocMany(byte, width * height);
	lwlib_PicViewer_MemToLatch(grsegs[pic], width, height, source);
	img = new Img();
	img->fromSource(source, width, height, 1, width, true, gamepal);
#ifdef SDL_TRUE_COLOR
	img->expandPalette(false);
#endif

    UNCACHEGRCHUNK(pic);

	return img;
}

void PicViewer::setGl(Gl *gl)
{
	priv.gl = gl;
}

void PicViewer::setKeys(lwlib_Keys_t keys)
{
	priv.keys = keys;
}

void PicViewer::setFont(Font *font)
{
	priv.font = font;
}

void PicViewer::setPalette(SDL_Color *palette)
{
	priv.palette = palette;
}

void PicViewer::setEdit(Edit *edit)
{
	priv.edit = edit;
}
