/*************************************************************************
** Wolf3D Legacy
** Copyright (C) 2012 by LinuxWolf - Team RayCAST
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**************************************************************************
** LinuxWolf Library for Wolf3D Legacy
*************************************************************************/

#include "lw.h"
using namespace lwlib;

#if 0
static lwlib_Raster_Vertex_t lwlib_SDLGlEmu_RasterVertex(
	lwlib_SDLGlEmu_t sdlGlEmu, double x, double y, double z, double u, 
	double v)
{
	lwlib_Raster_Vertex_t rasterVertex;

	const lwlib_TPoint4f pt = lwlib_vec4f(x, y, z, 1.0);
	const lwlib_TPoint4f ptXform = lwlib_mat4f_xform(priv.curMat, pt);

	const double iz = 1 / lwlib_Z(ptXform);

	lwlib_Zero(rasterVertex);
	rasterVertex.x = (int)lwlib_X(ptXform);
	rasterVertex.y = (int)lwlib_Y(ptXform);
	rasterVertex.uiz = u * iz;
	rasterVertex.viz = v * iz;
	rasterVertex.iz = iz;

	return rasterVertex;
}
#endif

lwlib_Raster_Vertex_t SDLGlEmu::rasterVertexFast2D(double x,
	double y, double u, double v)
{
	lwlib_Raster_Vertex_t rasterVertex;

	const Point4f pt = vec4f(x, y, 1.0, 1.0);
	const Point4f ptXform = mat_xform(priv.curMat, pt);

	lwlib_Zero(rasterVertex);
	rasterVertex.x = (int)lwlib_X(ptXform);
	rasterVertex.y = (int)lwlib_Y(ptXform);
	rasterVertex.uiz = u;
	rasterVertex.viz = v;

	return rasterVertex;
}

static lwlib_Byte_t *lwlib_SDLGlEmu_LockSurface(SDL_Surface *surface)
{
    if(SDL_MUSTLOCK(surface))
    {
        if(SDL_LockSurface(surface) < 0)
            return NULL;
    }
    return (lwlib_Byte_t *) surface->pixels;
}

static void lwlib_SDLGlEmu_UnlockSurface(SDL_Surface *surface)
{
    if(SDL_MUSTLOCK(surface))
    {
        SDL_UnlockSurface(surface);
    }
}

SDLGlEmu::SDLGlEmu()
{
	lwlib_Zero(priv);

	priv.curMat = mat4f_ident(4, 4);

	const lwlib_Color_Pix_t blankPix = 0;
	priv.blankTexId = uploadTex((lwlib_Byte_t *)&blankPix, 1, 1, 
		sizeof(lwlib_Color_Pix_t), NULL);
}

SDLGlEmu::~SDLGlEmu()
{
	int i;
	SDLGlEmu_Texture_t *texture;

	free(priv.depthBuffer);
	for (i = 0; (size_t)i < vec_size(priv.textureVec); i++)
	{
		texture = &vec_at(priv.textureVec, i);
		delete texture->img;
	}
	vec_free(priv.textureVec, SDLGlEmu_Texture_t);
}

void SDLGlEmu::setVideoMode(SDLGlEmu_Screen_t *screen)
{
	bool isDoubleBuffered;
	SDL_Surface *screenSurf;
#ifndef SDL_TRUE_COLOR
	SDL_Surface *screenBuffer;
#endif
	Uint32 sdlFlags;

#ifdef SDL_TRUE_COLOR
	if (screen->bits != 32)
	{
		printf("Only RGBA support is available\n");
		return;
	}

	sdlFlags = 
		(screen->useDoubleBuffering ? SDL_HWSURFACE | SDL_DOUBLEBUF : 0) |
		(screen->fullscreen ? SDL_FULLSCREEN : 0);
	printf("Setting video mode %ix%ix%i %s\n", screen->width, screen->height, 
		screen->bits, screen->useDoubleBuffering ? "(double buffered)" : "");
#else
	sdlFlags = SDL_SWSURFACE | (screen->bits == 8 ? SDL_HWPALETTE : 0) | 
		(screen->fullscreen ? SDL_FULLSCREEN : 0);
	printf("Setting video mode %ix%ix%i\n", screen->width, screen->height, 
		screen->bits);
#endif

	screenSurf = SDL_SetVideoMode(screen->width, screen->height, screen->bits,
		sdlFlags);
	if (screenSurf == NULL)
	{
		printf("Unable to set video mode\n");
		return;
	}
#ifdef SDL_TRUE_COLOR
	if (screenSurf->pitch % 4 != 0)
	{
		printf("Unexpected screen pitch %i\n", screenSurf->pitch);
		return;
	}
	priv.screenPitch = screenSurf->pitch / 4;
#else
	SDL_SetColors(screenSurf, screen->palette, 0, 256);
	priv.screenPitch = screenSurf->pitch;

	screenBuffer = SDL_CreateRGBSurface(SDL_SWSURFACE, screen->width,
		screen->height, 8, 0, 0, 0, 0);
	if(!screenBuffer)
	{
		printf("Unable to create screen buffer surface: %s\n", SDL_GetError());
		exit(1);
	}

	priv.screenBuffer = screenBuffer;
	SDL_SetColors(screenBuffer, screen->palette, 0, 256);

	priv.bufferPitch = screenBuffer->pitch;
#endif

	isDoubleBuffered = ((screenSurf->flags & SDL_DOUBLEBUF) == SDL_DOUBLEBUF);

	screen->screenSurf = screenSurf;
	screen->isDoubleBuffered = isDoubleBuffered;

	priv.screenSurf = screenSurf;
	priv.screenWidth = screen->width;
	priv.screenHeight = screen->height;
	priv.isDoubleBuffered = isDoubleBuffered;
	priv.depthBuffer = lwlib_CallocMany(fixedpt, screen->width * 
		screen->height);
#ifndef SDL_TRUE_COLOR
	priv.palette = screen->palette;
#endif
}

void SDLGlEmu::init()
{
}

void SDLGlEmu::clearScreen()
{
	lwlib_Color_Pix_t col;

#ifdef SDL_TRUE_COLOR
	col = SDL_MapRGBA(priv.screenSurf->format, 0, 0, 0, 0);
	SDL_FillRect(priv.screenSurf, NULL, col);
#else
	col = 0;
	SDL_FillRect(priv.screenBuffer, NULL, col);
#endif

	memset(priv.depthBuffer, 0, 
		priv.screenWidth * priv.screenHeight * sizeof(fixedpt));
}

void SDLGlEmu::updateScreen()
{

#ifdef SDL_TRUE_COLOR
	SDL_Flip(priv.screenSurf);
#else
	SDL_BlitSurface(priv.screenBuffer, NULL, priv.screenSurf, NULL);
	SDL_UpdateRect(priv.screenSurf, 0, 0, 0, 0);
#endif
}

int SDLGlEmu::uploadTex(lwlib_Byte_t *data, int w, int h, int bpp, 
	SDL_Color *palette)
{
	int i;
	Img_t img;
	SDLGlEmu_Texture_t texture;

	assert(bpp == 1 || bpp == 3 || bpp == 4);

	img = new Img();
	img->fromSource(data, w, h, bpp, w * bpp, false, palette);
	img->copyData();
	img->expandDimsPow2();
	if (bpp != 1)
	{
		img->expandBits(4);
		img->swapRGB();
	}

	lwlib_Zero(texture);
	texture.img = img;

	for (i = 0; (size_t)i < vec_size(priv.textureVec); i++)
	{
		if (vec_at(priv.textureVec, i).freed)
		{
			vec_at(priv.textureVec, i) = texture;
			break;
		}
	}
	if ((size_t)i == vec_size(priv.textureVec))
	{
		vec_push(priv.textureVec, SDLGlEmu_Texture_t, texture);
	}

	const int texId = i + 1;
	return texId;
}

void SDLGlEmu::deleteTex(int texId)
{
	SDLGlEmu_Texture_t *texture;

	if (texId <= 0)
	{
		return;
	}
	const int i = texId - 1;

	if ((size_t)i == vec_size(priv.textureVec) - 1)
	{
		vec_pop_back(priv.textureVec);
	}
	else
	{
		texture = &vec_at(priv.textureVec, i);
		delete texture->img;
		texture->freed = true;
	}
}

void SDLGlEmu::lockScreen()
{
	SDL_Surface *surf;

#ifdef SDL_TRUE_COLOR
	surf = priv.screenSurf;
#else
	surf = priv.screenBuffer;
#endif
	priv.vbuf = (lwlib_Color_Pix_t *)lwlib_SDLGlEmu_LockSurface(
		surf);
}

void SDLGlEmu::unlockScreen()
{
	SDL_Surface *surf;

#ifdef SDL_TRUE_COLOR
	surf = priv.screenSurf;
#else
	surf = priv.screenBuffer;
#endif
	lwlib_SDLGlEmu_UnlockSurface(surf);
}

void SDLGlEmu::drawScaledImage(int texId, 
	int ix, int iy, int iw, int ih, int w, int h, int cx, int cy, 
	int sw, int sh, bool transparent)
{
	Img_t img;
	SDLGlEmu_Texture_t *texture;
	lwlib_Raster_Screen_t *rasterScreen;
	lwlib_Raster_Viewport_t *rasterViewport;
	lwlib_Raster_Texture_t *rasterTexture;
	lwlib_Raster_Vertex_t rasterVertexs[2];
	lwlib_Raster_DrawPoly_t rasterDrawPoly;

	if (texId <= 0)
	{
		return;
	}
	const int textureIndex = texId - 1;

	lwlib_Zero(rasterDrawPoly);

	rasterScreen = &rasterDrawPoly.screen;
	rasterScreen->vbuf = priv.vbuf;
#ifdef SDL_TRUE_COLOR
	rasterScreen->pitch = priv.screenPitch;
#else
	rasterScreen->pitch = priv.bufferPitch;
#endif
	rasterScreen->zbuf = priv.depthBuffer;

	rasterViewport = &rasterDrawPoly.viewport;
	rasterViewport->x = 0;
	rasterViewport->y = 0;
	rasterViewport->w = priv.screenWidth;
	rasterViewport->h = priv.screenHeight;

	texture = &vec_at(priv.textureVec, textureIndex);
	rasterTexture = &rasterDrawPoly.texture;

	img = texture->img;
	rasterTexture->tex = (lwlib_Color_Pix_t *)img->getData();
	rasterTexture->nu = img->getHeight();
	rasterTexture->nv = img->getWidth();
	rasterTexture->nuLog2 = img->getHeightShift();
	rasterTexture->nvLog2 = img->getWidthShift();

	const int u1 = iy * rasterTexture->nu / h;
	const int v1 = ix * rasterTexture->nv / w;
	const int u2 = (iy + ih) * rasterTexture->nu / h;
	const int v2 = (ix + iw) * rasterTexture->nv / w;

	rasterVertexs[0] = rasterVertexFast2D(cx, cy, u1, v1);
	rasterVertexs[1] = rasterVertexFast2D(cx + sw, cy + sh, u2, v2);

	rasterDrawPoly.vertexs = rasterVertexs;
	rasterDrawPoly.blendPix = priv.curColor;
	rasterDrawPoly.affine = true;
	rasterDrawPoly.fast2D = true;
	lwlib_Raster_DrawPoly(&rasterDrawPoly);
}

void SDLGlEmu::setCurColor(Point4b colorRgba)
{
#ifdef SDL_TRUE_COLOR
	priv.curColor = LWLIB_COLOR_PACK(lwlib_X(colorRgba), lwlib_Y(colorRgba), 
		lwlib_Z(colorRgba), lwlib_W(colorRgba));
#else
	priv.curColor = lwlib_Img_FindClosestColor(priv.palette, 
		lwlib_X(colorRgba), lwlib_Y(colorRgba), lwlib_Z(colorRgba));
#endif
}

void SDLGlEmu::resetCurColor()
{
	setCurColor(vec4b(255, 255, 255, 255));
}

void SDLGlEmu::drawBox(int x, int y, int w, int h, Point4b colorRgba, 
	bool lines)
{
	lwlib_Color_Pix_t *data;
	SDLGlEmu_Texture_t *texture;
	const bool transparent = (lwlib_W(colorRgba) != 255);

#ifdef SDL_TRUE_COLOR
	const lwlib_Color_Pix_t pix = LWLIB_COLOR_PACK(lwlib_X(colorRgba), 
		lwlib_Y(colorRgba), lwlib_Z(colorRgba), lwlib_W(colorRgba));
#else
	const lwlib_Color_Pix_t pix = lwlib_Img_FindClosestColor(priv.palette, 
		lwlib_X(colorRgba), lwlib_Y(colorRgba), lwlib_Z(colorRgba));
#endif
	assert(priv.blankTexId > 0);
	texture = &vec_at(priv.textureVec, priv.blankTexId - 1);
	data = (lwlib_Color_Pix_t *)texture->img->getData();
	*data = pix;

	if (!lines)
	{
		drawScaledImage(priv.blankTexId, 0, 0, 1, 1, 1, 1, x, y, w, h, 
			transparent);
	}
	else
	{
		drawScaledImage(priv.blankTexId, 0, 0, 1, 1, 1, 1, x, y, w, 1, 
			transparent);
		drawScaledImage(priv.blankTexId, 0, 0, 1, 1, 1, 1, x + w - 1, y, 
			1, h, transparent);
		drawScaledImage(priv.blankTexId, 0, 0, 1, 1, 1, 1, x, y + h - 1, 
			w, 1, transparent);
		drawScaledImage(priv.blankTexId, 0, 0, 1, 1, 1, 1, x, y, 1, h, 
			transparent);
	}
}

void SDLGlEmu::drawBoxArrow(int x, int y, int w, int h, int r, 
	Point4b colorRgba, double dir)
{
}

void SDLGlEmu::pushMatrix()
{
	if (priv.matStackSize < LWLIB_SDLGLEMU_MATSTACK_MAX)
	{
		priv.matStack[priv.matStackSize++] = priv.curMat;
	}
}

void SDLGlEmu::popMatrix()
{
	if (priv.matStackSize > 0)
	{
		priv.curMat = priv.matStack[--priv.matStackSize];
	}
}

void SDLGlEmu::translateMatrix(double tx, double ty, double tz)
{
	Mat4f mat;

	mat = mat4f_ident(4, 4);
	lwlib_CM4(mat, 0, 3) = tx;
	lwlib_CM4(mat, 1, 3) = ty;
	lwlib_CM4(mat, 2, 3) = tz;

	priv.curMat = mat_mult(mat, priv.curMat);
}

void SDLGlEmu::drawLine(int x1, int y1, int x2, int y2, Point4b colorRgba)
{
}

int SDLGlEmu::getWidth()
{
	return priv.screenWidth;
}

int SDLGlEmu::getHeight()
{
	return priv.screenHeight;
}

Point2i SDLGlEmu::getDims()
{
	return vec2i(priv.screenWidth, priv.screenHeight);
}
