/*************************************************************************
** Wolf3D Legacy
** Copyright (C) 2012 by LinuxWolf - Team RayCAST
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**************************************************************************
** LinuxWolf Library for Wolf3D Legacy
*************************************************************************/

#include <SDL.h>
#include "lw_timer.h"
#include "lw_misc.h"
using namespace lwlib;

Timer lwlib::g_frameTimer;

Timer::Timer()
{
	lwlib_Zero(priv.ms);
	priv.ms = SDL_GetTicks();
}

bool Timer::elapsed(uint32_t ms) const
{
	return (Timer().priv.ms >= priv.ms + ms);
}

double Timer::gapInSeconds(const Timer &a) const
{
	return ((double)a.priv.ms - (double)priv.ms) / 1000.0;
}

void Timer::addDuration(double sec)
{
	priv.ms += (uint32_t)(sec * 1000.0);
}

double Timer::secondsElapsed() const
{
	if (priv.ms == 0)
	{
		return 0.0;
	}

	return gapInSeconds(Timer());
}

TimerFps::TimerFps()
{
	lwlib_Zero(priv);
}

void TimerFps::updateFrame()
{
	float currFps;

	priv.oldTime = priv.newTime;
	priv.newTime = SDL_GetTicks();

	priv.frames++;

	if (priv.newTime < priv.lastFpsTime)
	{
		priv.lastFpsTime = priv.newTime;
	}

	if ((priv.newTime - priv.lastFpsTime) > 1000)
	{
		float seconds = (priv.newTime - priv.lastFpsTime) / 1000.0;
		currFps = priv.frames / seconds;
		priv.lastFpsTime = priv.newTime;
		priv.frames = 0;
		priv.fpsShow++;
	}

	if (priv.frames == 0 && priv.fpsShow == 5)
	{
		printf("%g FPS\n", currFps);
		priv.fpsShow = 0;
	}
}
