// WL_ACT1.C

#include "wl_def.h"
#pragma hdrstop

/*
=============================================================================

                                                        STATICS

=============================================================================
*/


statobj_t       statobjlist[MAXSTATS];
statobj_t       *laststatobj;

/*
===============
=
= InitStaticList
=
===============
*/

void InitStaticList (void)
{
    laststatobj = &statobjlist[0];
    memset(statat, 0xff, sizeof(statat));
}



/*
===============
=
= SpawnStatic
=
===============
*/

void SpawnStatic (int tilex, int tiley, int type)
{
    SpawnSpecialStatic(tilex, tiley, ED_STATINFO[type].type,
        ED_SPR(ED_STATINFO[type].picnum), ED_STATINFO[type].specialFlags,
        ED_STATINFO[type].animframes, ED_STATINFO[type].bonus, type);
}

void SpawnSpecialStatic (int tilex, int tiley, int type, short picnum, 
    uint32_t specialFlags, byte animframes, const ED_BonusDef_t bonusDef,
    int statInfoNum)
{
    ED_Bonus_t *bonus;
    laststatobj->shapenum = picnum;
    laststatobj->tilex = tilex;
    laststatobj->tiley = tiley;
    laststatobj->visspot = &spotvis[tilex][tiley];
    laststatobj->animframes = animframes;
    laststatobj->itemnumber = type;
    laststatobj->bonusIndex = ED_BonusIndex(ed, bonusDef);
    laststatobj->statInfoNum = statInfoNum;

    if (laststatobj->bonusIndex >= 0)
    {
        laststatobj->flags = FL_BONUS;

        bonus = &ED_Bonuses(ed)[laststatobj->bonusIndex];
        if (bonus->isTreasure)
        {
            if (!loadedgame)
            {
                gamestate.treasuretotal++;
            }
        }

        if (bonus->usePickup)
        {
            actorat[tilex][tiley] = (objtype *) BLOCKTILE;
            specialFlags |= FL_BLOCK;
        }
    }

    switch (type)
    {
        case block:
            actorat[tilex][tiley] = (objtype *) BLOCKTILE;          // consider it a blocking tile
        case none:
        case bo_mg42_mounted:
            laststatobj->flags = 0;
            break;
        case bo_any:
            break;
    }

    if (specialFlags)
        laststatobj->flags |= specialFlags;

    statat[tilex][tiley] = laststatobj - statobjlist;
    laststatobj++;

    if (laststatobj == &statobjlist[MAXSTATS])
        Quit ("Too many static objects!\n");
}


/*
===============
=
= PlaceItemType
=
= Called during game play to drop actors' items.  It finds the proper
= item number based on the item type (bo_???).  If there are no free item
= spots, nothing is done.
=
===============
*/

void PlaceItemType (const ED_BonusDef_t bonusDef, int tilex, int tiley)
{
    int type;
    statobj_t *spot;
    int statObjNum;

    // no item drop on mounted weapon
    statObjNum = statat[tilex][tiley];
    spot = statObjNum != -1 ? &statobjlist[statObjNum] : NULL;
    if (spot != NULL && spot->itemnumber == bo_mg42_mounted)
    {
        return;
    }

    //
    // find the item number
    //
    for (type=0; ; type++)
    {
        if (ED_SPR(ED_STATINFO[type].picnum) == -1)                    // end of list
            Quit ("PlaceItemType: couldn't find type!");
        if (strcmp(ED_STATINFO[type].bonus, bonusDef) == 0)
            break;
    }

    //
    // find a spot in statobjlist to put it in
    //
    for (spot=&statobjlist[0]; ; spot++)
    {
        if (spot==laststatobj)
        {
            if (spot == &statobjlist[MAXSTATS])
                return;                                     // no free spots
            laststatobj++;                                  // space at end
            break;
        }

        if (spot->shapenum == -1)                           // -1 is a free spot
            break;
    }
    //
    // place it
    //
    spot->shapenum = ED_SPR(ED_STATINFO[type].picnum);
    spot->tilex = tilex;
    spot->tiley = tiley;
    spot->visspot = &spotvis[tilex][tiley];
    spot->animframes = ED_STATINFO[type].animframes;
    if (ED_STATINFO[type].type == bo_mg42_mounted)
        actorat[tilex][tiley] = (objtype *) BLOCKTILE;          // consider it a blocking tile
    else
        spot->flags = FL_BONUS | ED_STATINFO[type].specialFlags;
    spot->itemnumber = ED_STATINFO[type].type;
    spot->bonusIndex = ED_BonusIndex(ed, bonusDef);
    spot->statInfoNum = type;
    statat[tilex][tiley] = spot - statobjlist;
}


/*
===============
=
= RemoveStaticObj
=
= Removes static object from list by putting special mark on it
=
===============
*/
void RemoveStaticObj(statobj_t *check)
{
    statat[check->tilex][check->tiley] = -1;
    check->shapenum = -1;               // remove from list
    if (check->itemnumber == block)
    {
        actorat[check->tilex][check->tiley] = NULL;
    }
}

/*
=============================================================================

                                  DOORS

doorobjlist[] holds most of the information for the doors

doorposition[] holds the amount the door is open, ranging from 0 to 0xffff
        this is directly accessed by AsmRefresh during rendering

The number of doors is limited to 64 because a spot in tilemap holds the
        door number in the low 6 bits, with the high bit meaning a door center
        and bit 6 meaning a door side tile

Open doors conect two areas, so sounds will travel between them and sight
        will be checked when the player is in a connected area.

Areaconnect is incremented/decremented by each door. If >0 they connect

Every time a door opens or closes the areabyplayer matrix gets recalculated.
        An area is true if it connects with the player's current spor.

=============================================================================
*/

#define DOORWIDTH       0x7800
#define OPENTICS        300

doorobj_t       doorobjlist[MAXDOORS],*lastdoorobj;
short           doornum;

word            doorposition[MAXDOORS];             // leading edge of door 0=closed
                                                    // 0xffff = fully open

byte            areaconnect[NUMAREAS][NUMAREAS];

boolean         areabyplayer[NUMAREAS];

/*
==============
=
= ConnectAreas
=
= Scans outward from playerarea, marking all connected areas
=
==============
*/

void RecursiveConnect (int areanumber)
{
    int i;

    for (i=0;i<NUMAREAS;i++)
    {
        if (areaconnect[areanumber][i] && !areabyplayer[i])
        {
            areabyplayer[i] = true;
            RecursiveConnect (i);
        }
    }
}


void ConnectAreas (void)
{
    memset (areabyplayer,0,sizeof(areabyplayer));
    areabyplayer[player->areanumber] = true;
    RecursiveConnect (player->areanumber);
}


void InitAreas (void)
{
    memset (areabyplayer,0,sizeof(areabyplayer));
    if (player->areanumber < NUMAREAS)
        areabyplayer[player->areanumber] = true;
}


/*
==============
=
= AreaByArea
=
= WSJ: quick way of checking if two areas are connected
=
==============
*/

boolean AreaByArea (int area1, int area2)
{
    if (area1 >= NUMAREAS || area2 >= NUMAREAS || area1 < 0 || area2 < 0)
        return false;

    if (area1 == area2 || areaconnect[area1][area2] > 0 || areaconnect[area2][area1] > 0)
        return true;

    return false;
}


/*
===============
=
= InitDoorList
=
===============
*/

void InitDoorList (void)
{
    memset (areabyplayer,0,sizeof(areabyplayer));
    memset (areaconnect,0,sizeof(areaconnect));

    lastdoorobj = &doorobjlist[0];
    doornum = 0;
}


/*
===============
=
= SpawnDoor
=
===============
*/

void SpawnDoor (int tilex, int tiley, boolean vertical, int lock)
{
    word *map;

    if (doornum==MAXDOORS)
        Quit ("64+ doors on level!");

    doorposition[doornum] = 0;              // doors start out fully closed
    lastdoorobj->tilex = tilex;
    lastdoorobj->tiley = tiley;
    lastdoorobj->vertical = vertical;
    lastdoorobj->lock = lock;
    lastdoorobj->action = dr_closed;

    actorat[tilex][tiley] = (objtype *)(uintptr_t)(doornum | 0x80);   // consider it a solid wall

    //
    // make the door tile a special tile, and mark the adjacent tiles
    // for door sides
    //
    tilemap[tilex][tiley] = doornum | 0x80;
    map = mapsegs[0] + (tiley<<mapshift) +tilex;
    if (vertical)
    {
        *map = *(map-1);                        // set area number
    }
    else
    {
        *map = *(map-mapwidth);                 // set area number
    }

    doornum++;
    lastdoorobj++;
}

//===========================================================================

/*
=====================
=
= PlayDoorSound
=
= WSJ: play one door sound at a time
=
=====================
*/

void PlayDoorSound (int sound, int door)
{
    PlaySoundLocTile(sound,doorobjlist[door].tilex,doorobjlist[door].tiley);
    PlaySoundLocTile(sound,doorobjlist[door].tilex,doorobjlist[door].tiley);
}


/*
=====================
=
= OpenDoor
=
=====================
*/

void OpenDoor (int door)
{
    if (doorobjlist[door].action == dr_open)
        doorobjlist[door].ticcount = 0;         // reset open time
    else
    {
        int area = *(mapsegs[0] + (doorobjlist[door].tiley<<mapshift)
            +doorobjlist[door].tilex)-AREATILE;
        if (doorobjlist[door].action == dr_closing && areabyplayer[area])
        {
            // WSJ: door was re-opened
            PlayDoorSound(ED_GetThemedSound(ed, "OPENDOORSND"),door);
        }
        doorobjlist[door].action = dr_opening;  // start it opening
    }
}


/*
=====================
=
= CloseDoor
=
=====================
*/

void CloseDoor (int door)
{
    int     tilex,tiley,area;
    objtype *check;

    //
    // don't close on anything solid
    //
    tilex = doorobjlist[door].tilex;
    tiley = doorobjlist[door].tiley;

    if (actorat[tilex][tiley])
        return;

    if (player->tilex == tilex && player->tiley == tiley)
        return;

    if (doorobjlist[door].vertical)
    {
        if ( player->tiley == tiley )
        {
            if ( ((OB_GetX(player) + MINDIST) >> TILESHIFT) == tilex )
                return;
            if ( ((OB_GetX(player)-MINDIST) >>TILESHIFT) == tilex )
                return;
        }
        check = actorat[tilex-1][tiley];
        if (ISPOINTER(check) && ((OB_GetX(check)+MINDIST) >> TILESHIFT) == tilex )
            return;
        check = actorat[tilex+1][tiley];
        if (ISPOINTER(check) && ((OB_GetX(check)-MINDIST) >> TILESHIFT) == tilex )
            return;
    }
    else if (!doorobjlist[door].vertical)
    {
        if (player->tilex == tilex)
        {
            if ( ((OB_GetY(player)+MINDIST) >>TILESHIFT) == tiley )
                return;
            if ( ((OB_GetY(player)-MINDIST) >>TILESHIFT) == tiley )
                return;
        }
        check = actorat[tilex][tiley-1];
        if (ISPOINTER(check) && ((OB_GetY(check)+MINDIST) >> TILESHIFT) == tiley )
            return;
        check = actorat[tilex][tiley+1];
        if (ISPOINTER(check) && ((OB_GetY(check)-MINDIST) >> TILESHIFT) == tiley )
            return;
    }


    //
    // play door sound if in a connected area
    //
    area = *(mapsegs[0] + (doorobjlist[door].tiley<<mapshift)
        +doorobjlist[door].tilex)-AREATILE;
    if (areabyplayer[area])
    {
        PlayDoorSound(ED_GetThemedSound(ed, "CLOSEDOORSND"),door);
    }

    doorobjlist[door].action = dr_closing;
    //
    // make the door space solid
    //
    actorat[tilex][tiley] = (objtype *)(uintptr_t)(door | 0x80);
}



/*
=====================
=
= CheckLock
=
=====================
*/

boolean CheckLock (int door)
{
    int lock = doorobjlist[door].lock;
    int tilex = doorobjlist[door].tilex;
    int tiley = doorobjlist[door].tiley;

    switch (lock)
    {
        case dr_lock1:
            if (!(gamestate.keys & 1))
            {
                PlayDoorSound(ED_SND("DOORLOCKEDSND"),door);
                GetMessage("You need the Gold Key to open this door");
                return false;
            }
            break;
        case dr_lock2:
            if (!(gamestate.keys & 2))
            {
                PlayDoorSound(ED_SND("DOORLOCKEDSND"),door);
                GetMessage("You need the Silver Key to open this door");
                return false;
            }
            break;
        case dr_lock3:
            if (!(gamestate.keys & 4))
            {
                PlayDoorSound(ED_SND("DOORLOCKEDSND"),door);
                GetMessage("You need the Bronze Key to open this door");
                return false;
            }
            break;
        case dr_lock4:
            if (!(gamestate.keys & 8))
            {
                PlayDoorSound(ED_SND("DOORLOCKEDSND"),door);
                GetMessage("You need the Turquois Key to open this door");
                return false;
            }
            break;
        case dr_lock_fw:
            if (player->tilex < tilex || player->tiley < tiley)
            {
                PlayDoorSound(ED_SND("NOWAYSND"),door);
                GetMessage("This door is locked from the other side");
                return false;
            }
            break;
        case dr_lock_bw:
            if (player->tilex > tilex || player->tiley > tiley)
            {
                PlayDoorSound(ED_SND("NOWAYSND"),door);
                GetMessage("This door is locked from the other side");
                return false;
            }
            break;
    }

    return true;
}

void OperateDoor (int door)
{
    switch (doorobjlist[door].action)
    {
        case dr_closed:
            if (!CheckLock(door))
                return;
        case dr_closing:
            OpenDoor (door);
            break;
        case dr_open:
        case dr_opening:
            CloseDoor (door);
            break;
    }
}


//===========================================================================

/*
===============
=
= DoorOpen
=
= Close the door after three seconds
=
===============
*/

void DoorOpen (int door)
{
    int opentime;

    opentime = OPENTICS;

    if ( (doorobjlist[door].ticcount += (short) tics) >= opentime)
        CloseDoor (door);
}



/*
===============
=
= DoorOpening
=
===============
*/

void DoorOpening (int door)
{
    unsigned area1,area2;
    word *map;
    int32_t position;

    position = doorposition[door];
    if (!position)
    {
        //
        // door is just starting to open, so connect the areas
        //
        map = mapsegs[0] + (doorobjlist[door].tiley<<mapshift)
            +doorobjlist[door].tilex;

        if (doorobjlist[door].vertical)
        {
            area1 = *(map+1);
            area2 = *(map-1);
        }
        else
        {
            area1 = *(map-mapwidth);
            area2 = *(map+mapwidth);
        }
        area1 -= AREATILE;
        area2 -= AREATILE;

        if (area1 < NUMAREAS && area2 < NUMAREAS)
        {
            areaconnect[area1][area2]++;
            areaconnect[area2][area1]++;

            if (player->areanumber < NUMAREAS)
                ConnectAreas ();

            if (areabyplayer[area1])
                PlayDoorSound(ED_GetThemedSound(ed, "OPENDOORSND"),door);
        }
    }

    //
    // slide the door by an adaptive amount
    //
    position += tics<<10;
    if (position >= 0xffff)
    {
        //
        // door is all the way open
        //
        position = 0xffff;
        doorobjlist[door].ticcount = 0;
        doorobjlist[door].action = dr_open;
        actorat[doorobjlist[door].tilex][doorobjlist[door].tiley] = 0;
    }

    doorposition[door] = (word) position;
}


/*
===============
=
= DoorClosing
=
===============
*/

void DoorClosing (int door)
{
    unsigned area1,area2;
    word *map;
    int32_t position;
    int tilex,tiley;

    tilex = doorobjlist[door].tilex;
    tiley = doorobjlist[door].tiley;

    if ( ((int)(uintptr_t)actorat[tilex][tiley] != (door | 0x80))
        || (player->tilex == tilex && player->tiley == tiley) )
    {                       // something got inside the door
        OpenDoor (door);
        return;
    };

    position = doorposition[door];

    //
    // slide the door by an adaptive amount
    //
    position -= tics<<10;
    if (position <= 0)
    {
        //
        // door is closed all the way, so disconnect the areas
        //
        position = 0;

        doorobjlist[door].action = dr_closed;

        map = mapsegs[0] + (doorobjlist[door].tiley<<mapshift) + doorobjlist[door].tilex;

        if (doorobjlist[door].vertical)
        {
            area1 = *(map+1);
            area2 = *(map-1);
        }
        else
        {
            area1 = *(map-mapwidth);
            area2 = *(map+mapwidth);
        }
        area1 -= AREATILE;
        area2 -= AREATILE;

        if (area1 < NUMAREAS && area2 < NUMAREAS)
        {
            areaconnect[area1][area2]--;
            areaconnect[area2][area1]--;

            if (player->areanumber < NUMAREAS)
                ConnectAreas ();
        }
    }

    doorposition[door] = (word) position;
}




/*
=====================
=
= MoveDoors
=
= Called from PlayLoop
=
=====================
*/

void MoveDoors (void)
{
    int door;

    if (gamestate.victoryflag)              // don't move door during victory sequence
        return;

    for (door = 0; door < doornum; door++)
    {
        switch (doorobjlist[door].action)
        {
            case dr_open:
                DoorOpen (door);
                break;

            case dr_opening:
                DoorOpening(door);
                break;

            case dr_closing:
                DoorClosing(door);
                break;

            default:
                break;
        }
    }
}


/*
=============================================================================

                                PUSHABLE WALLS

=============================================================================
*/

word pwallstate;
word pwallpos;                  // amount a pushable wall has been moved (0-63)
word pwallx,pwally;
byte pwalldir,pwalltile,pwallpatch;
int dirs[4][2]={{0,-1},{1,0},{0,1},{-1,0}};

/*
===============
=
= PushWall
=
===============
*/

void PushWall (int checkx, int checky, int dir)
{
    int oldtile, oldpatch, dx, dy;

    if (pwallstate)
        return;

    oldtile = tilemap[checkx][checky];
    oldpatch = wallpatch[checkx][checky];
    if (!oldtile)
        return;

    dx = dirs[dir][0];
    dy = dirs[dir][1];

    if (actorat[checkx+dx][checky+dy])
    {
        //SD_PlaySound (NOWAYSND);
        return;
    }
    actorat[checkx+dx][checky+dy] = (objtype *)(uintptr_t) (tilemap[checkx+dx][checky+dy] = oldtile);
    wallpatch[checkx+dx][checky+dy] = oldpatch;

    gamestate.secretcount++;
    pwallx = checkx;
    pwally = checky;
    pwalldir = dir;
    pwallstate = 1;
    pwallpos = 0;
    pwalltile = tilemap[pwallx][pwally];
    pwallpatch = wallpatch[pwallx][pwally];
    tilemap[pwallx][pwally] = 64;
    tilemap[pwallx+dx][pwally+dy] = 64;
    wallpatch[pwallx][pwally] = 0;
    *(mapsegs[1]+(pwally<<mapshift)+pwallx) = 0;   // remove P tile info
    *(mapsegs[0]+(pwally<<mapshift)+pwallx) = *(mapsegs[0]+(player->tiley<<mapshift)+player->tilex); // set correct floorcode (BrotherTank's fix)

    PlaySoundLocTile(ED_SND("PUSHWALLSND"),pwallx,pwally);
    GetMessage("You found a secret!");
}



/*
=================
=
= MovePWalls
=
=================
*/

void MovePWalls (void)
{
    int oldblock,oldtile,oldpatch;

    if (!pwallstate)
        return;

    oldblock = pwallstate/128;

    pwallstate += (word)tics;

    if (pwallstate/128 != oldblock)
    {
        // block crossed into a new block
        oldtile = pwalltile;
        oldpatch = pwallpatch;

        //
        // the tile can now be walked into
        //
        tilemap[pwallx][pwally] = 0;
        wallpatch[pwallx][pwally] = 0;
        actorat[pwallx][pwally] = 0;
        *(mapsegs[0]+(pwally<<mapshift)+pwallx) = player->areanumber+AREATILE;

        int dx=dirs[pwalldir][0], dy=dirs[pwalldir][1];
        //
        // see if it should be pushed farther
        //
        if (pwallstate>=512)            // only move two tiles fix
        {
            //
            // the block has been pushed two tiles
            //
            pwallstate = 0;
            tilemap[pwallx+dx][pwally+dy] = oldtile;
            wallpatch[pwallx+dx][pwally+dy] = oldpatch;
            return;
        }
        else
        {
            int xl,yl,xh,yh;
            xl = (OB_GetX(player)-PLAYERSIZE) >> TILESHIFT;
            yl = (OB_GetY(player)-PLAYERSIZE) >> TILESHIFT;
            xh = (OB_GetX(player)+PLAYERSIZE) >> TILESHIFT;
            yh = (OB_GetY(player)+PLAYERSIZE) >> TILESHIFT;

            pwallx += dx;
            pwally += dy;

            if (actorat[pwallx+dx][pwally+dy]
                || (xl<=pwallx+dx && pwallx+dx<=xh && 
                    yl<=pwally+dy && pwally+dy<=yh))
            {
                pwallstate = 0;
                tilemap[pwallx][pwally] = oldtile;
                wallpatch[pwallx][pwally] = oldpatch;
                return;
            }
            actorat[pwallx+dx][pwally+dy] = (objtype *)(uintptr_t) (tilemap[pwallx+dx][pwally+dy] = oldtile);
            tilemap[pwallx+dx][pwally+dy] = 64;
            wallpatch[pwallx+dx][pwally+dy] = 0;
            PlaySoundLocTile(ED_SND("PUSHWALLSND"),pwallx,pwally);
        }
    }

    pwallpos = (pwallstate/2)&63;
}

/*
=================
=
= GetMapSpot
=
=================
*/

word GetMapSpot(int x, int y, int plane)
{
    if (plane == 0 && (tileflags[x][y] & TILEFLAGS_DESTROYED) != 0)
    {
        return 0;
    }

    return mapsegs[plane][(y << mapshift) + x];
}


/*
=================
=
= DestroyWall
=
=================
*/

void DestroyWall(int x, int y)
{
    if (tilemap[x][y] != 0)
    {
        actorat[x][y] = NULL;
        tileflags[x][y] |= TILEFLAGS_DESTROYED;
        tilemap[x][y] = 0;
    }
}
