// WL_GOAL.CPP

#include <stdio.h>
#include <math.h>
#include "wl_def.h"
#include "lw.h"
#if defined _WIN32
#pragma hdrstop
#endif

/*
===================================================================

                           GOAL OBJECTS

===================================================================
*/

bool OB_CanTargetObj(objtype *ob, objtype *check)
{
    int i;
    const char *goalEnemyName;
    const ED_Enemy_t *goalEnemy;

    if (OB_CanResurrect(ob))
    {
        return OB_CanTargetObjResurrect(ob, check);
    }

    for (i = 0; (size_t)i < vec_size(ob->enemy->goalEnemyNames); i++)
    {
        goalEnemyName = vec_at(ob->enemy->goalEnemyNames, i).v;
        goalEnemy = ED_Enemy(ed, goalEnemyName);

        if (check->enemy != goalEnemy)
        {
            // goal obj is not of the right type for this enemy
            continue;
        }

        if (check->hitpoints <= 0)
        {
            // goal object has no health
            continue;
        }

        if (OB_TileDist(check, player) < OB_TileDist(check, ob))
        {
            // player is closer to this goal object than enemy
            continue;
        }

        return true;
    }

    return false;
}

void OB_ChooseGoalObj(objtype *ob)
{
    objtype *check;
    objtype *goalObj;
    objtype *foundObj;
    int foundObjIndex = 0;
    int minTileDist = MAPSIZE;

    if (!OB_CanResurrect(ob))
    {
        if (vec_size(ob->enemy->goalEnemyNames) == 0)
        {
            return;
        }
    }

    if (ob->areanumber != player->areanumber ||
        ob->reaction == react_findmg42)
    {
        return;
    }

    for (ITEROBJ(check, 1))
    {
        if (!OB_CanTargetObj(ob, check))
        {
            continue;
        }

        goalObj = check;

        // allow souls to target bodies in areas not occupied by the
        // player
        if (OB_CanResurrect(ob))
        {
            if (goalObj->areanumber != ob->areanumber)
            {
                // this goal object is not in same area
                continue;
            }
        }
        // unlike mutants who can only target a mecha suit in the same
        // area as the player
        else
        {
            if (goalObj->areanumber != player->areanumber)
            {
                // this goal object is not in same area as player
                continue;
            }
        }

        if (goalObj->objIndexArr[oit_goalseeker] != 0)
        {
            // this goal object is already targeted by another actor
            continue;
        }

        if (!OB_CheckCoordLine(ob, goalObj))
        {
            // enemy doesn't have line of sight to this goal object
            continue;
        }

        // find the closest goal object available
        if (OB_TileDist(goalObj, ob) < minTileDist)
        {
            foundObjIndex = goalObj - objlist;
            minTileDist = OB_TileDist(goalObj, ob);
        }
    }

    if (foundObjIndex != 0)
    {
        foundObj = &objlist[foundObjIndex];

        ob->objIndexArr[oit_goaltarget] = foundObj - objlist;
        foundObj->objIndexArr[oit_goalseeker] = ob - objlist;

        ob->reaction = react_findmg42;
    }
}

bool OB_CanUseGoalObj(objtype *ob, int x, int y)
{
    objtype *check;

    check = actorat[x][y];

    return ob->reaction == react_findmg42 && 
        ISPOINTER(check) &&
        ob->objIndexArr[oit_goaltarget] == check - objlist;
}
