#include "version.h"

#ifdef USE_PARALLAX

#include "wl_def.h"
#include "wl_shade.h"

#ifdef USE_FEATUREFLAGS

// The lower left tile of every map determines the start texture of the parallax sky.
static int GetParallaxStartTexture()
{
    int startTex = ffDataBottomLeft;
    assert(startTex >= 0 && startTex < PMSpriteStart);
    return startTex;
}

#else

static int GetParallaxStartTexture()
{
    int startTex;
    switch(gamestate.episode * 10 + mapon)
    {
        case  0: skyID = 20; break;
        default: skyID =  0; break;
    }
    assert(startTex >= 0 && startTex < PMSpriteStart);
    return startTex;
}

#endif

void DrawParallax(byte *vbuf, unsigned vbufPitch)
{
    int startpage = GetParallaxStartTexture();
    int midangle = Mecha_GetViewingFineAngle();
    int skyheight = viewheight >> 1;
    int curtex = -1;
    byte *skytex;
#ifdef USE_SHADING
    byte *curshades;
#ifdef USE_FEATUREFLAGS
    if(!lightningtics && GetFeatureFlags() & FF_SHADESKY)
        curshades = shadetable[ShadeGlow::off][(SHADE_COUNT/2)];
    else
#endif
        curshades = shadetable[ShadeGlow::off][0];
#endif

    startpage += USE_PARALLAX - 1;

    for(int x = 0; x < viewwidth; x++)
    {
        int curang = pixelangle[x] + midangle;
        if(curang < 0) curang += FINEANGLES;
        else if(curang >= FINEANGLES) curang -= FINEANGLES;
        int xtex = curang * USE_PARALLAX * TEXTURESIZE / FINEANGLES;
        int newtex = xtex >> TEXTURESHIFT;
        if(newtex != curtex)
        {
            curtex = newtex;
            skytex = PM_GetTexture(startpage - curtex);
        }
        int texoffs = TEXTUREMASK - ((xtex & (TEXTURESIZE - 1)) << TEXTURESHIFT);
        //int yend = skyheight - (wallheight[x] >> 3);
        int yend = skyheight;
        if(yend <= 0) continue;

        for(int y = 0, offs = x; y < yend; y++, offs += vbufPitch)
#ifdef USE_SHADING
            vbuf[offs] = curshades[skytex[texoffs + (y * TEXTURESIZE) / skyheight]];
#else
            vbuf[offs] = skytex[texoffs + (y * TEXTURESIZE) / skyheight];
#endif
    }
}

#endif
