// WL_PLAY.C

#include "wl_def.h"
#pragma hdrstop

#include "wl_cloudsky.h"
#include "wl_shade.h"
#include "lw.h"

/*
=============================================================================

                                                 LOCAL CONSTANTS

=============================================================================
*/

#define sc_Question     0x35

/*
=============================================================================

                                                 GLOBAL VARIABLES

=============================================================================
*/

boolean madenoise;              // true when shooting or screaming
byte    alertedarea,lastalertedarea;

exit_t playstate;

static musicnames lastmusicchunk = (musicnames) 0;

static int DebugOk;

objtype objlist[MAXACTORS];
int objcount = 0;
int objlistend = 0;

objtype *newobj, *obj, *player, *lastobj, *objfreelist, *killerobj;

boolean noclip, ammocheat;
#ifdef GODMODE_DEFAULTON
int godmode = 1;
#else
int godmode;
#endif
int invisiblemode;
int singlestep;
int extravbls = 1; // to remove flicker (gray stuff at the bottom)

byte tilemap[MAPSIZE][MAPSIZE];    // wall values only
byte wallpatch[MAPSIZE][MAPSIZE];  // WSJ: wall patch values
flat_t flatmap[MAPSIZE][MAPSIZE];
byte spotvis[MAPSIZE][MAPSIZE];
objtype *actorat[MAPSIZE][MAPSIZE];
int statat[MAPSIZE][MAPSIZE];
int ambiance[NUMAREAS];    // WSJ: set during SetupGameLevel
byte tileflags[MAPSIZE][MAPSIZE];  // wall flags

//
// replacing refresh manager
//
unsigned tics;
unsigned lightningtics;    // WSJ: amount of time everything is lit up during a lightning flash

//
// control info
//
boolean mouseenabled, joystickenabled, alwaysrun, aimassist;
int dirscan[4] = { sc_UpArrow, sc_RightArrow, sc_DownArrow, sc_LeftArrow };
int buttonscan[NUMBUTTONS][2] =
{ 
	{ sc_Control, sc_None }, // bt_attack
	{ sc_X, sc_None }, // bt_altattack
	{ sc_Alt, sc_None }, // bt_strafe
	{ sc_LShift, sc_None }, // bt_run
	{ sc_Space, sc_Enter }, // bt_use
	{ sc_R, sc_None }, // bt_reload
	{ sc_BackSpace, sc_M }, // bt_usemedkit
	{ sc_1, sc_None }, // bt_readyknife
	{ sc_2, sc_None }, // bt_readypistol
	{ sc_3, sc_None }, // bt_readyrifle
	{ sc_4, sc_None }, // bt_readymachinegun
	{ sc_5, sc_None }, // bt_readyshotgun
	{ sc_6, sc_None }, // bt_readyflamethrower
	{ sc_7, sc_None }, // bt_readychaingun
	{ sc_8, sc_None }, // bt_readylauncher
	{ sc_N, sc_None }, // bt_nextweapon
	{ sc_P, sc_None }, // bt_prevweapon
	{ sc_Escape, sc_None }, // bt_esc
	{ sc_Pause, sc_None }, // bt_pause
	{ sc_Comma, sc_A }, // bt_strafeleft
	{ sc_Period, sc_D }, // bt_straferight
	{ sc_UpArrow, sc_W }, // bt_moveforward
	{ sc_DownArrow, sc_S }, // bt_movebackward
	{ sc_LeftArrow, sc_None }, // bt_turnleft
	{ sc_RightArrow, sc_None }, // bt_turnright
};
int buttonmouse[5] =
{ 
    bt_attack, bt_strafe, bt_use, bt_nextweapon, 
    bt_prevweapon
};
int buttonjoy[32] = {
#ifdef _arch_dreamcast
    bt_attack, bt_strafe, bt_use, bt_run, bt_esc, bt_prevweapon, bt_nobutton, bt_nextweapon,
    bt_pause, bt_strafeleft, bt_straferight, bt_nobutton, bt_nobutton, bt_nobutton, bt_nobutton, bt_nobutton,
#else
    bt_attack, bt_strafe, bt_use, bt_run, bt_strafeleft, bt_straferight, bt_esc, bt_pause,
    bt_prevweapon, bt_nextweapon, bt_nobutton, bt_nobutton, bt_nobutton, bt_nobutton, bt_nobutton, bt_nobutton,
#endif
    bt_nobutton, bt_nobutton, bt_nobutton, bt_nobutton, bt_nobutton, bt_nobutton, bt_nobutton, bt_nobutton,
    bt_nobutton, bt_nobutton, bt_nobutton, bt_nobutton, bt_nobutton, bt_nobutton, bt_nobutton, bt_nobutton
};

int viewsize;

boolean buttonheld[NUMBUTTONS];

boolean demorecord, demoplayback;
int8_t *demoptr, *lastdemoptr;
memptr demobuffer;

//
// current user input
//
int controlx, controly;         // range from -100 to 100 per tic
boolean buttonstate[NUMBUTTONS];

int lastgamemusicoffset = 0;


//===========================================================================


void CenterWindow (word w, word h);
void InitObjList (void);
void RemoveObj (objtype * gone);
void PollControls (void);
int StopMusic (void);
void StartMusic (void);
void ContinueMusic (int offs);
void PlayLoop (void);

/*
=============================================================================

                                                 LOCAL VARIABLES

=============================================================================
*/


objtype dummyobj;


/*
=============================================================================

                               USER CONTROL

=============================================================================
*/

/*
===================
=
= PollKeyboardButtons
=
===================
*/

void PollKeyboardButtons (void)
{
    int i;

    for (i = 0; i < NUMBUTTONS; i++)
    {
        if (Keyboard[buttonscan[i][0]] || Keyboard[buttonscan[i][1]])
            buttonstate[i] = true;
    }
}


/*
===================
=
= PollMouseButtons
=
===================
*/

void PollMouseButtons (void)
{
    int buttons = IN_MouseButtons ();

    if (buttons & 1)
        buttonstate[buttonmouse[0]] = true;
    if (buttons & 2)
        buttonstate[buttonmouse[1]] = true;
    if (buttons & 4)
        buttonstate[buttonmouse[2]] = true;
    if (buttons & 8)
        buttonstate[buttonmouse[3]] = true;
    if (buttons & 16)
        buttonstate[buttonmouse[4]] = true;
}



/*
===================
=
= PollJoystickButtons
=
===================
*/

void PollJoystickButtons (void)
{
    int buttons = IN_JoyButtons();

    for(int i = 0, val = 1; i < JoyNumButtons; i++, val <<= 1)
    {
        if(buttons & val)
            buttonstate[buttonjoy[i]] = true;
    }
}


/*
===================
=
= PollKeyboardMove
=
===================
*/

void PollKeyboardMove (void)
{
    int delta;
    mecha_t *mecha;

    mecha = &gamestate.mecha;

    if (Mecha_Equipped() && !mecha->lightWeight)
    {
        delta = BASEMOVE * tics;
    }
    else
    {
        delta = (alwaysrun ? !buttonstate[bt_run] : buttonstate[bt_run]) ? 
            RUNMOVE * tics : BASEMOVE * tics;
    }

    if (buttonstate[bt_moveforward])
        controly -= delta;
    if (buttonstate[bt_movebackward])
        controly += delta;
    if (buttonstate[bt_turnleft])
        controlx -= delta;
    if (buttonstate[bt_turnright])
        controlx += delta;
}


/*
===================
=
= PollMouseMove
=
===================
*/

void PollMouseMove (void)
{
    int mousexmove, mouseymove;

    SDL_GetMouseState(&mousexmove, &mouseymove);
    if(IN_IsInputGrabbed())
        IN_CenterMouse();

    mousexmove -= screenWidth / 2;
    mouseymove -= screenHeight / 2;

    if (mouseadjustment >= 10)
    {
        controlx += mousexmove * 20;
    }
    else
    {
        controlx += mousexmove * 10 / (10 - mouseadjustment);
    }
    controly += mouseymove * 20 / (13 - mouseadjustment);
}


/*
===================
=
= PollJoystickMove
=
===================
*/

void PollJoystickMove (void)
{
    int joyx, joyy;

    IN_GetJoyDelta (&joyx, &joyy);

    int delta = buttonstate[bt_run] ? RUNMOVE * tics : BASEMOVE * tics;

    if (joyx > 64 || buttonstate[bt_turnright])
        controlx += delta;
    else if (joyx < -64  || buttonstate[bt_turnleft])
        controlx -= delta;
    if (joyy > 64 || buttonstate[bt_movebackward])
        controly += delta;
    else if (joyy < -64 || buttonstate[bt_moveforward])
        controly -= delta;
}

/*
===================
=
= PollControls
=
= Gets user or demo input, call once each frame
=
= controlx              set between -100 and 100 per tic
= controly
= buttonheld[]  the state of the buttons LAST frame
= buttonstate[] the state of the buttons THIS frame
=
===================
*/

void PollControls (void)
{
    int max, min, i;
    byte buttonbits;

    IN_ProcessEvents();

//
// get timing info for last frame
//
    if (demoplayback || demorecord)   // demo recording and playback needs to be constant
    {
        // wait up to DEMOTICS Wolf tics
        uint32_t curtime = SDL_GetTicks();
        lasttimecount += DEMOTICS;
        int32_t timediff = (lasttimecount * 100) / 7 - curtime;
        if(timediff > 0)
            SDL_Delay(timediff);

        if(timediff < -2 * DEMOTICS)       // more than 2-times DEMOTICS behind?
            lasttimecount = (curtime * 7) / 100;    // yes, set to current timecount

        tics = DEMOTICS;
    }
    else
        CalcTics ();

    controlx = 0;
    controly = 0;
    memcpy (buttonheld, buttonstate, sizeof (buttonstate));
    memset (buttonstate, 0, sizeof (buttonstate));

    if (demoplayback)
    {
        //
        // read commands from demo buffer
        //
        buttonbits = *demoptr++;
        for (i = 0; i < NUMBUTTONS; i++)
        {
            buttonstate[i] = buttonbits & 1;
            buttonbits >>= 1;
        }

        controlx = *demoptr++;
        controly = *demoptr++;

        if (demoptr == lastdemoptr)
            playstate = ex_completed;   // demo is done

        controlx *= (int) tics;
        controly *= (int) tics;

        return;
    }


//
// get button states
//
    PollKeyboardButtons ();

    if (mouseenabled && IN_IsInputGrabbed())
        PollMouseButtons ();

    if (joystickenabled)
        PollJoystickButtons ();

//
// get movements
//
    PollKeyboardMove ();

    if (mouseenabled && IN_IsInputGrabbed())
        PollMouseMove ();

    if (joystickenabled)
        PollJoystickMove ();

//
// bound movement to a maximum
//
    max = 100 * tics;
    min = -max;
    if (controlx > max)
        controlx = max;
    else if (controlx < min)
        controlx = min;

    if (controly > max)
        controly = max;
    else if (controly < min)
        controly = min;

    if (demorecord)
    {
        //
        // save info out to demo buffer
        //
        controlx /= (int) tics;
        controly /= (int) tics;

        buttonbits = 0;

        // TODO: Support 32-bit buttonbits
        for (i = NUMBUTTONS - 1; i >= 0; i--)
        {
            buttonbits <<= 1;
            if (buttonstate[i])
                buttonbits |= 1;
        }

        *demoptr++ = buttonbits;
        *demoptr++ = controlx;
        *demoptr++ = controly;

        if (demoptr >= lastdemoptr - 8)
            playstate = ex_completed;
        else
        {
            controlx *= (int) tics;
            controly *= (int) tics;
        }
    }
}



//==========================================================================



///////////////////////////////////////////////////////////////////////////
//
//      CenterWindow() - Generates a window of a given width & height in the
//              middle of the screen
//
///////////////////////////////////////////////////////////////////////////
#define MAXX    320
#define MAXY    160

void CenterWindow (word w, word h)
{
    US_DrawWindow (((MAXX / 8) - w) / 2, ((MAXY / 8) - h) / 2, w, h);
}

//===========================================================================


/*
=====================
=
= CheckKeys
=
=====================
*/

void CheckKeys (void)
{
    ScanCode scan;
    int i;
    lwlib_String_t message;


    if (screenfaded || demoplayback)    // don't do anything with a faded screen
        return;

    scan = LastScan;

    if (Keyboard[sc_U])
    {
        int kr, sr, tr;

        memset(message, '\0', sizeof(message));

        kr = sr = tr = 100;
        if (gamestate.killtotal)
            kr = (gamestate.killcount * 100) / gamestate.killtotal;
        if (gamestate.secrettotal)
            sr = (gamestate.secretcount * 100) / gamestate.secrettotal;
        if (gamestate.treasuretotal)
            tr = (gamestate.treasurecount * 100) / gamestate.treasuretotal;

        lwlib_Snprintf(message, "Kill Ratio: %d%%\nSecret Ratio: %d%%\n"
            "Treasure Ratio: %d%%", kr, sr, tr);

        GetMessage (message);
    }

#ifdef DEBUGKEYS
    //
    // SECRET CHEAT CODE: TAB-G-F10
    //
    if (Keyboard[sc_Tab] && Keyboard[sc_G] && Keyboard[sc_F10])
    {
        WindowH = 160;
        if (godmode)
        {
            Message ("God mode OFF");
            SD_PlaySound (ED_SND("NOBONUSSND"));
        }
        else
        {
            Message ("God mode ON");
            SD_PlaySound (ED_SND("ENDBONUS2SND"));
        }

        IN_Ack ();
        godmode ^= 1;
        ammocheat ^= 1;
        DrawPlayBorderSides ();
        IN_ClearKeysDown ();
        return;
    }

    //
    // SECRET CHEAT CODE: 'RIP'
    //
    if (param_debugmode && Keyboard[sc_R] && Keyboard[sc_I] && 
        Keyboard[sc_P])
    {
        gamestate.health = 100;
        gamestate.armor = 200;
        //gamestate.gotregenarmor = gamestate.gotregenhealth = true;
        for (i=0; i<NUMAMMOTYPES; i++)
        {
            gamestate.maxammoindex = NUMAMMOLIMITS - 1;
            gamestate.ammo[i] = CURMAXAMMO(i);
        }
        gamestate.keys = 15;
        gamestate.score = 0;
        gamestate.medkits = MAX_MEDKITS;
        gamestate.medkitlim = MAX_MEDKITS;
        gamestate.TimeCount += 42000L;
        for (i=0; i<NUMWEAPONS; i++) { GiveWeapon((weapontype)i); }
        DrawWeapon ();
        DrawHealth ();
        DrawKeys ();
        DrawAmmo ();
        DrawScore ();
        DrawLives ();

        ClearMemory ();
        CA_CacheGrChunk (STARTFONT + 1);
        ClearSplitVWB ();

        Message (STR_CHEATER1 "\n"
                 STR_CHEATER2 "\n\n" STR_CHEATER3 "\n" STR_CHEATER4 "\n" STR_CHEATER5);

        UNCACHEGRCHUNK (STARTFONT + 1);
        IN_ClearKeysDown ();
        IN_Ack ();

        if (viewsize < 17)
            DrawPlayBorder ();
    }

    if ((buttonstate[bt_usemedkit] && !buttonheld[bt_usemedkit]))
    {
        if(!gamestate.medkits || gamestate.health > 99)
        {
            GetMessage(lwlib_VaString(message, 
                "Portable Medkits: %d", gamestate.medkits));
        }
        else
        {
            gamestate.medkits--;
            HealSelf (25);
            DrawPlayScreen();
            GetMessage ("Used a portable medkit.");
            SD_PlaySound (ED_SND("HEALTH2SND"));
        }
    }

    //
    // OPEN UP DEBUG KEYS
    //
    if (Keyboard[sc_BackSpace] && Keyboard[sc_LShift] && Keyboard[sc_Alt] && param_debugmode)
    {
        ClearMemory ();
        CA_CacheGrChunk (STARTFONT + 1);
        ClearSplitVWB ();

        Message ("Debugging keys are\nnow available!");
        UNCACHEGRCHUNK (STARTFONT + 1);
        IN_ClearKeysDown ();
        IN_Ack ();

        DrawPlayBorderSides ();
        DebugOk = 1;
    }
#endif

//
// pause key weirdness can't be checked as a scan code
//
    if(buttonstate[bt_pause]) Paused = true;
    if(Paused)
    {
        int lastoffs = StopMusic();
        LatchDrawPic (20 - 4, 80 - 2 * 8, PAUSEDPIC);
        VH_UpdateScreen();
        IN_Ack ();
        Paused = false;
        ContinueMusic(lastoffs);
        if (MousePresent && IN_IsInputGrabbed())
            IN_CenterMouse();     // Clear accumulated mouse movement
        lasttimecount = GetTimeCount();
        return;
    }

//
// F1-F7/ESC to enter control panel
//
    if (CP_CheckQuick (scan))
    {
        ClearMemory ();
        ED_ResetAmbients(ed);
        ClearSplitVWB ();

        DrawPlayBorderSides ();

        SETFONTCOLOR (0, 15);
        IN_ClearKeysDown ();
        return;
    }

    if ((scan >= sc_F1 && scan <= sc_F9) || scan == sc_Escape || buttonstate[bt_esc])
    {
        int lastoffs = StopMusic ();
        ClearMemory ();
        ED_ResetAmbients(ed);
        VW_FadeOut ();

        US_ControlPanel (buttonstate[bt_esc] ? sc_Escape : scan);

        SETFONTCOLOR (0, 15);
        IN_ClearKeysDown ();
        VW_FadeOut();
        if(viewsize != 21)
            DrawPlayScreen ();
        if (!startgame && !loadedgame)
            ContinueMusic (lastoffs);
        if (loadedgame)
            playstate = ex_abort;
        lasttimecount = GetTimeCount();
        if (MousePresent && IN_IsInputGrabbed())
            IN_CenterMouse();     // Clear accumulated mouse movement
        return;
    }

//
// TAB-? debug keys
//
#ifdef DEBUGKEYS
    if (Keyboard[sc_Tab] && DebugOk)
    {
        CA_CacheGrChunk (STARTFONT);
        fontnumber = 0;
        SETFONTCOLOR (0, 15);
        if (DebugKeys () && viewsize < 18)
            DrawPlayBorder ();       // dont let the blue borders flash

        if (MousePresent && IN_IsInputGrabbed())
            IN_CenterMouse();     // Clear accumulated mouse movement

        lasttimecount = GetTimeCount();
        return;
    }
#endif
}


//===========================================================================

/*
#############################################################################

                                  The objlist data structure

#############################################################################

objlist containt structures for every actor currently playing.  The structure
is accessed as a linked list starting at *player, ending when ob->next ==
NULL.  GetNewObj inserts a new object at the end of the list, meaning that
if an actor spawn another actor, the new one WILL get to think and react the
same frame.  RemoveObj unlinks the given object and returns it to the free
list, but does not damage the objects ->next pointer, so if the current object
removes itself, a linked list following loop can still safely get to the
next element.

<backwardly linked free list>

#############################################################################
*/


/*
=========================
=
= InitActorList
=
= Call to clear out the actor object lists returning them all to the free
= list.  Allocates a special spot for the player.
=
=========================
*/

void InitActorList (void)
{
//
// init the actor lists
//
    lwlib_Zero(objlist);
    objcount = objlistend = 0;

//
// give the player the first free spots
//
    GetNewActor ();
    player = newobj;
}

//===========================================================================

/*
=========================
=
= GetNewActor
=
= Sets the global variable new to point to a free spot in objlist.
= The free spot is inserted at the end of the liked list
=
= When the object list is full, the caller can either have it bomb out ot
= return a dummy object pointer that will never get used
=
=========================
*/

void GetNewActor (void)
{
    objtype *ob;

    if (objcount >= MAXACTORS)
    {
        Quit ("GetNewActor: Too many objects!");
    }

#ifdef DEBUG_OBJECT_MANAGER
    printf("PRE object add: objlistend=%d, objcount=%d\n", objlistend, objcount);
#endif

    for (ob = objlist; ob < objlist + objlistend; ob++)
    {
        if (ob->alloced == 0)
        {
            break;
        }
    }

    if (ob == objlist + objlistend)
    {
        if (objlistend >= MAXACTORS)
        {
            Quit ("GetNewActor: No free spots in objlist!");
        }
        objlistend++;
    }
    objcount++;

    newobj = ob;
    memset (newobj, 0, sizeof (*newobj));
    newobj->alloced = 1;
    newobj->active = ac_no;
    newobj->enemy = ED_EnemyByIndex(ed, newobj->enemyIndex);
    newobj->proj = ED_ProjByIndex(ed, newobj->projIndex);
    newobj->state = ED_StateByIndex(ed, newobj->stateIndex);

#ifdef DEBUG_OBJECT_MANAGER
    printf("POST object add: objlistend=%d, objcount=%d\n", objlistend, objcount);
#endif
}

//===========================================================================

#ifdef DEBUG_CHECK_OBJINDEXARR
void CheckObjIndexArr(void)
{
    int i;
    objtype *ob;

    for (i = 0; i < objlistend; i++)
    {
        ob = &objlist[i];
        if (ob->alloced == 0)
        {
            continue;
        }

        if (ob->objIndexArr[oit_goaltarget] != 0)
        {
            assert(objlist[ob->objIndexArr[oit_goaltarget]].objIndexArr
                [oit_goalseeker] == ob - objlist);
        }
        if (ob->objIndexArr[oit_goalseeker] != 0)
        {
            assert(objlist[ob->objIndexArr[oit_goalseeker]].objIndexArr
                [oit_goaltarget] == ob - objlist);
        }
    }
}
#endif

/*
=========================
=
= RemoveObj
=
= Add the given object back into the free list, and unlink it from it's
= neighbors
=
=========================
*/

void RemoveObj (objtype * gone)
{
    int objIndex;

    if (gone == player)
    {
        Quit ("RemoveObj: Tried to remove the player!");
    }

    if (objcount <= 0)
    {
        Quit ("RemoveObj: No objects left to remove!");
    }

    if (gone - objlist < 0 || gone - objlist >= MAXACTORS)
    {
        Quit ("RemoveObj: Tried to remove object out of memory!");
    }

    if (gone - objlist >= objlistend)
    {
        Quit ("RemoveObj: Tried to remove object beyond end of list!");
    }

    ResetGoal(gone);

    objIndex = gone->objIndexArr[oit_goalseeker];
    if (objIndex != 0)
    {
        ResetGoal(&objlist[objIndex]);
    }

    if (actorat[gone->tilex][gone->tiley] == gone)
    {
        actorat[gone->tilex][gone->tiley] = NULL;
    }

#ifdef DEBUG_OBJECT_MANAGER
    printf("PRE object remove: gone=%d, objlistend=%d, objcount=%d\n",
        gone - objlist, objlistend, objcount);
#endif

    lwlib_Zero(*gone);

    if (gone - objlist == objlistend - 1)
    {
        while (gone >= objlist && gone->alloced == 0)
        {
            gone--;
        }
        objlistend = (gone - objlist) + 1;
    }
    objcount--;

#ifdef DEBUG_OBJECT_MANAGER
    printf("POST object remove: gone=%d, objlistend=%d, objcount=%d\n",
        gone - objlist, objlistend, objcount);
#endif

#ifdef DEBUG_CHECK_OBJINDEXARR
    CheckObjIndexArr();
#endif
}

void IterObjReset(objtype **ob, int c)
{
    *ob = objlist;
    while (*ob - objlist < objlistend)
    {
        if ((*ob)->alloced == 1)
        {
            break;
        }
        (*ob)++;
    }

    while (c--)
    {
        IterObjNext(ob);
    }
}

bool IterObjMore(objtype *ob)
{
    return ob - objlist < objlistend && ob->alloced == 1;
}

void IterObjNext(objtype **ob)
{
    while ((*ob) - objlist < objlistend)
    {
        (*ob)++;
        if ((*ob)->alloced == 1)
        {
            break;
        }
    }
}

void MoveObjToFreeSpot(objtype *ob, objtype *freeob)
{
    int objIndex;
    objtype *seeker, *target;

    if (actorat[ob->tilex][ob->tiley] == ob)
    {
        actorat[ob->tilex][ob->tiley] = freeob;
    }

    objIndex = ob->objIndexArr[oit_goalseeker];
    if (objIndex != 0)
    {
        seeker = &objlist[objIndex];
        assert(seeker->objIndexArr[oit_goaltarget] == ob - objlist);
        seeker->objIndexArr[oit_goaltarget] = freeob - objlist;
    }

    objIndex = ob->objIndexArr[oit_goaltarget];
    if (objIndex != 0)
    {
        target = &objlist[objIndex];
        assert(target->objIndexArr[oit_goalseeker] == ob - objlist);
        target->objIndexArr[oit_goalseeker] = freeob - objlist;
    }
    
    *freeob = *ob;
    lwlib_Zero(*ob);
}

/*
=============================================================================

                                                MUSIC STUFF

=============================================================================
*/


/*
=================
=
= StopMusic
=
=================
*/
int StopMusic (void)
{
    int lastoffs = SD_MusicOff ();

    UNCACHEAUDIOCHUNK (STARTMUSIC + lastmusicchunk);

    return lastoffs;
}

//==========================================================================


/*
=================
=
= StartMusic
=
=================
*/

void StartMusic ()
{
    ContinueMusic(0);
}

/*
=================
=
= ContinueMusic
=
=================
*/

void ContinueMusic (int offs)
{
    int trackNum;

    const int bossTrackNum = gamestate.bossmusictracknum;

    trackNum = (gamestate.activebosses > 0 ? 
        (bossTrackNum > 0 ? bossTrackNum : ffDataBossMusic) :
        ffDataMusic);

    SD_MusicOff ();
    lastmusicchunk = (musicnames) (BLANK_MUS+trackNum);
    SD_ContinueMusic(STARTMUSIC + lastmusicchunk, offs);
}

/*
=================
=
= UpdateBossMusic
=
=================
*/

void UpdateBossMusic(objtype *ob, bool newActiveBoss)
{
    ED_Music_t *music;

    if (newActiveBoss)
    {
        gamestate.bossmusictracknum = 0;

        music = ob->enemy->music;
        if (music != NULL && music->mapNum == gamestate.mapon)
        {
            gamestate.bossmusictracknum = music->trackNum;
        }

        gamestate.activebosses++;
        if (gamestate.activebosses == 1)
        {
            StartMusic();
        }
    }
    else
    {
        if (gamestate.activebosses > 0)
        {
            gamestate.activebosses--;
            if (gamestate.activebosses == 0)
            {
                StartMusic();
            }
        }
    }
}

/*
=============================================================================

                                        PALETTE SHIFTING STUFF

=============================================================================
*/

#define NUMREDSHIFTS    6
#define REDSTEPS        8

#define NUMWHITESHIFTS  3
#define WHITESTEPS      20
#define WHITETICS       6

SDL_Color redshifts[NUMREDSHIFTS][256];
SDL_Color whiteshifts[NUMWHITESHIFTS][256];
SDL_Color greenshifts[NUMREDSHIFTS][256];

int damagecount, bonuscount;
boolean palshifted;

/*
=====================
=
= InitRedShifts
=
=====================
*/

void InitRedShifts (void)
{
    SDL_Color *workptr, *baseptr;
    int i, j, delta;


//
// fade through intermediate frames
//
    for (i = 1; i <= NUMREDSHIFTS; i++)
    {
        workptr = redshifts[i - 1];
        baseptr = gamepal;

        for (j = 0; j <= 255; j++)
        {
            delta = 256 - baseptr->r;
            workptr->r = baseptr->r + delta * i / REDSTEPS;
            delta = -baseptr->g;
            workptr->g = baseptr->g + delta * i / REDSTEPS;
            delta = -baseptr->b;
            workptr->b = baseptr->b + delta * i / REDSTEPS;
            baseptr++;
            workptr++;
        }
    }

    for (i = 1; i <= NUMWHITESHIFTS; i++)
    {
        workptr = whiteshifts[i - 1];
        baseptr = gamepal;

        for (j = 0; j <= 255; j++)
        {
            delta = 256 - baseptr->r;
            workptr->r = baseptr->r + delta * i / WHITESTEPS;
            delta = 248 - baseptr->g;
            workptr->g = baseptr->g + delta * i / WHITESTEPS;
            delta = 0-baseptr->b;
            workptr->b = baseptr->b + delta * i / WHITESTEPS;
            baseptr++;
            workptr++;
        }
    }

    for (i = 1; i <= NUMREDSHIFTS; i++)
    {
        workptr = greenshifts[i - 1];
        baseptr = gamepal;

        for (j = 0; j <= 255; j++)
        {
            delta = -baseptr->r;
            workptr->r = baseptr->r + delta * i / REDSTEPS;
            delta = 256 - baseptr->g;
            workptr->g = baseptr->g + delta * i / REDSTEPS;
            delta = -baseptr->b;
            workptr->b = baseptr->b + delta * i / REDSTEPS;
            baseptr++;
            workptr++;
        }
    }
}


/*
=====================
=
= ClearPaletteShifts
=
=====================
*/

void ClearPaletteShifts (void)
{
    bonuscount = damagecount = 0;
    palshifted = false;
}


/*
=====================
=
= StartBonusFlash
=
=====================
*/

void StartBonusFlash (void)
{
    bonuscount = NUMWHITESHIFTS * WHITETICS;    // white shift palette
}


/*
=====================
=
= StartDamageFlash
=
=====================
*/

void StartDamageFlash (int damage)
{
    damagecount += damage;
}


/*
=====================
=
= UpdatePaletteShifts
=
=====================
*/

void UpdatePaletteShifts (void)
{
    int red, white;

    if (bonuscount)
    {
        white = bonuscount / WHITETICS + 1;
        if (white > NUMWHITESHIFTS)
            white = NUMWHITESHIFTS;
        bonuscount -= tics;
        if (bonuscount < 0)
            bonuscount = 0;
    }
    else
        white = 0;


    if (damagecount)
    {
        red = damagecount / 10 + 1;
        if (red > NUMREDSHIFTS)
            red = NUMREDSHIFTS;

        damagecount -= tics;
        if (damagecount < 0)
            damagecount = 0;
    }
    else
        red = 0;

    if (gamestate.infected)
    {
        red = MIN(red + 1, NUMREDSHIFTS);
        VL_SetPalette (greenshifts[red - 1]);
        palshifted = true;
    }
    else if (red)
    {
        VL_SetPalette (redshifts[red - 1]);
        palshifted = true;
    }
    else if (white)
    {
        VL_SetPalette (whiteshifts[white - 1]);
        palshifted = true;
    }
    else if (palshifted)
    {
        VL_SetPalette (gamepal);        // back to normal
        palshifted = false;
    }
}


/*
=====================
=
= FinishPaletteShifts
=
= Resets palette to normal if needed
=
=====================
*/

void FinishPaletteShifts (void)
{
    if (palshifted)
    {
        palshifted = 0;
        VL_SetPalette (gamepal);
    }
}


/*
=============================================================================

                                                CORE PLAYLOOP

=============================================================================
*/


/*
=====================
=
= MakeNoise
=
=====================
*/

void MakeNoise (objtype *ob)
{
    alertedarea = ob->areanumber;
    if (ob == player)
    {
        madenoise = true;
        lastalertedarea = player->areanumber;
    }
}


/*
=====================
=
= CheckGoalTile
=
= WSJ: checks if static object is an MG42,
= and if there are any enemies nearby who
= can use it. If so, the object will be
= set as the closest enemy's goal.
=
=====================
*/

void CheckGoalTile (void)
{
    statobj_t    *statptr;
    objtype      *check,*goalseeker;
    word         distance,mindistance=0xFFFF;

    for (statptr=&statobjlist[0]; statptr!=laststatobj; statptr++)
    {
        if (statptr->itemnumber != bo_mg42_mounted)
            continue;

        if (player->tilex == statptr->tilex && player->tiley == statptr->tiley)
            continue;

        if (AreaAt(statptr->tilex,statptr->tiley)!=player->areanumber)
            continue;

        if (statptr->flags&FL_ATTACKMODE || statptr->flags&FL_FIRSTATTACK)
            continue;

        goalseeker = player;
        for (ITEROBJ(check, 0))
        {
            if (!CanUseMG42(check) || check->reaction >= react_dead)
                continue;

            distance = TILEDIST(check->tilex,check->tiley,statptr->tilex,statptr->tiley);
            if (distance > TILEDIST(check->tilex,check->tiley,player->tilex,player->tiley)
            || distance > TILEDIST(player->tilex,player->tiley,statptr->tilex,statptr->tiley))
                continue;

            if (!CheckCoordLine(OB_GetX(check),OB_GetY(check),TILECENTER(statptr->tilex),TILECENTER(statptr->tiley),false))
                continue;

            if (distance < mindistance)
            {
                mindistance = distance;
                goalseeker = check;
            }
        }

        if (goalseeker == player)
            continue;

        goalseeker->goalitemnum = (statptr - statobjlist) + 1;
        goalseeker->reaction = react_findmg42;
        statptr->flags |= FL_FIRSTATTACK;
    }
}


/*
=====================
=
= DoActor
=
=====================
*/

void DoActor (objtype * ob)
{
    void (*think) (objtype *);
    void (*action) (objtype *);
    statetype *state;

    if (!ob->state)
    {
        RemoveObj (ob);
        return;
    }

    if (!ob->active && !areabyplayer[ob->areanumber] && !AreaByActor(ob,lastalertedarea))
        return;

    if (!(ob->flags & (FL_NONMARK | FL_NEVERMARK)))
        actorat[ob->tilex][ob->tiley] = NULL;

//
// non transitional object
//

    if (!ob->ticcount)
    {
        think = ED_THINK(ob->state->think);
        if (think)
        {
            think (ob);
            if (!ob->state)
            {
                RemoveObj (ob);
                return;
            }
        }

        if (ob->flags & FL_NEVERMARK)
            return;

        if ((ob->flags & FL_NONMARK) && actorat[ob->tilex][ob->tiley])
            return;

        actorat[ob->tilex][ob->tiley] = ob;
        return;
    }

//
// transitional object
//
    ob->ticcount -= (short) tics;
    while (ob->ticcount <= 0)
    {
        state = ob->state;

        // end of state action
        action = ED_THINK(ob->state->action);
        if (action)
        {
            action (ob);
            if (!ob->state)
            {
                RemoveObj (ob);
                return;
            }
        }

        if (ob->state == state)
        {
            OB_SetState(ob, ED_STATE(ob->state->next));
        }

        if (!ob->state)
        {
            RemoveObj (ob);
            return;
        }

        if (!ob->state->tictime)
        {
            ob->ticcount = 0;
            goto think;
        }

        ob->ticcount += ob->state->tictime;
    }

think:
    //
    // think
    //
    think = ED_THINK(ob->state->think);
    if (think)
    {
        think (ob);
        if (!ob->state)
        {
            RemoveObj (ob);
            return;
        }
    }

    if (ob->flags & FL_NEVERMARK)
        return;

    if ((ob->flags & FL_NONMARK) && actorat[ob->tilex][ob->tiley])
        return;

    actorat[ob->tilex][ob->tiley] = ob;
}

//==========================================================================

/*
===============
=
= DoThunder
=
===============
*/

int32_t thunderticcount,thunderdelay;

static void DoThunder (void)
{
#if defined(USE_FEATUREFLAGS) && defined(USE_RAIN)
    if (lightningtics > 0)
        lightningtics--;
    if(GetFeatureFlags() & FF_THUNDER)
    {
        if (!thunderdelay)
        {
            // WSJ: maximum: 10 sec, minimum: 3 sec
            thunderdelay = (RANDINT(860) + 210);
        }
        thunderticcount += tics;
        if (thunderticcount > thunderdelay)
        {   // WSJ: play thunder sound and flash
            int sounds[4]={
                ED_SND("AMBTHUNDER1SND"),
                ED_SND("AMBTHUNDER2SND"),
                ED_SND("AMBTHUNDER3SND")
            };
            int thundervol = SD_MaxVolume();
            if (ffDataThunderVol > 0 && ffDataThunderVol <= 255)
            {
                thundervol = SD_MaxVolume() * ffDataThunderVol / 255;
            }
            thunderticcount = thunderdelay = 0;    // WSJ: reset timer
            SD_PlaySoundVolume(sounds[RANDINT(3)], thundervol);
            lightningtics = 20;
        }
    }
#endif
}

//==========================================================================

/*
===============
=
= ComputeAmbientVolTile
=
= Distance scaling effect for ambient sounds.
=
===============
*/

lwlib::Point2f ComputeAmbientVolTile(ED_Ambient_t *ambient, int x, int y,
    double visFactor)
{
    int i;
    int ambIndex;
    int area;
    double distSq, contrib, cosFactor;
    bool addContrib;
    int radius;
    ED_AmbientSettings_t *ambientSettings;
    lwlib::Point3f tileCentre, playerPos;
    lwlib::Point2f earVolumes;
    lwlib::Point2i soundLocs;

    ambientSettings = ED_AmbientSettings(ed);

    if (!TILEINMAP(x, y) || tilemap[x][y])
    {
        return lwlib::vec2f_zero();
    }

    addContrib = false;

    switch (ambient->type)
    {
    case amb_area:
        ambIndex = ambient - ED_Ambients(ed);
        area = AreaAt(x, y);
        addContrib = ((ambiance[area] | gamestate.ambiance[area]) & 
            (1 << ambIndex)) != 0;
        break;
    case amb_scenery:
        for (i = 0; (size_t)i < vec_size(ambient->actors); i++)
        {
            if (MAPSPOT(x, y, 1) == vec_at(ambient->actors, i))
            {
                addContrib = true;
                break;
            }
        }
        break;
    case amb_rain:
        if (GetFeatureFlags() & FF_RAIN)
        {
            if (TILEINRAIN(x, y))
            {
                addContrib = true;
            }
        }
        break;
    case amb_raincover:
        if (GetFeatureFlags() & FF_RAIN)
        {
            if (!TILEINRAIN(x, y))
            {
                addContrib = true;
            }
        }
        break;
    case amb_water:
        if (TILEINWATER(x, y))
        {
            addContrib = true;
        }
        break;
    }

    if (addContrib)
    {
        radius = ambientSettings->radius;

        playerPos = lwlib::vec3f(fixedpt_toconst(player->x),
            fixedpt_toconst(player->y), 0.5);
        tileCentre = lwlib::vec3f(x + 0.5, y + 0.5, 0.5);
        distSq = vec_dist_sq(playerPos, tileCentre);

        contrib = 
            lwlib_MAX
                (
                    1.0 - 
                    (
                        lwlib_MIN(distSq, lwlib_Sq(radius)) / 
                        lwlib_Sq(radius)
                    ),
                    0.0
                );
        //contrib = 1.0f;

        if (ambient->type != amb_scenery)
        {
            contrib *= ambientSettings->scale;
        }

        contrib *= visFactor;

        SetSoundLoc(fixedpt_rconst(tileCentre.v[0]),
            fixedpt_rconst(tileCentre.v[1]), -1);
        soundLocs = lwlib::vec2i(leftchannel, rightchannel);

        for (i = 0; i < 2; i++)
        {
            cosFactor = (((15 - soundLocs.v[i]) << 4) + 15) / 255.0;
            earVolumes.v[i] = contrib * cosFactor;
        }
    }
    else
    {
        earVolumes = lwlib::vec2f_zero();
    }

    return earVolumes;
}

/*
===============
=
= TileAllowsAmbiance
=
===============
*/

static bool TileAllowsAmbiance(int x, int y)
{
    return
        // tile is in map
        TILEINMAP(x, y) && 
        // and either
        (
            // tile is area tile in same area as player
            (
                !tilemap[x][y] && 
                AreaByArea(player->areanumber, AreaAt(x, y))
            ) ||
            // or tile is a non-closed door tile
            (
                (tilemap[x][y] & 0x80) != 0 &&
                doorposition[tilemap[x][y] & 0x7f] != 0
            )
        );
}

/*
===============
=
= PrintVisFactorMat
=
===============
*/

#if 0
static void PrintVisFactorMat(const ED_VisFactorMat_t &visFactorMat)
{
	int x, y, vf, r;
    ED_AmbientSettings_t *ambientSettings;

    ambientSettings = ED_AmbientSettings(ed);
    r = ambientSettings->radius;

    printf("vols:\n");
    for (y = 0; y < 2 * r; y++)
    {
        for (x = 0; x < 2 * r; x++)
        {
            if (player->tilex - r + x == ambientSettings->ox &&
                player->tiley - r + y == ambientSettings->oy)
            {
                printf("** ");
                continue;
            }
            vf = lwlib_CLIP((int)(
                mat_elem(visFactorMat, y, x) * 99), -99, 99);
            if (vf != 0)
            {
                printf("%03d ", vf);
            }
            else
            {
                printf("... ");
            }
        }
        printf("\n");
    }
}
#endif

/*
===============
=
= AmbientDoorConnected
=
===============
*/

static bool AmbientDoorConnected(int x, int y)
{
    int areaNum;
    int doorNum;
    doorobj_t *door;

    doorNum = tilemap[x][y] & 0x7f;
    door = &doorobjlist[doorNum];

    areaNum = player->areanumber;
    return door->vertical ?
        AreaByArea(areaNum, AreaAt(x - 1, y)) || 
        AreaByArea(areaNum, AreaAt(x + 1, y)) :
        AreaByArea(areaNum, AreaAt(x, y - 1)) || 
        AreaByArea(areaNum, AreaAt(x, y + 1));
}

/*
===============
=
= ComputeAmbientVisFactor
=
===============
*/

double ComputeAmbientVisFactor(int x, int y, int px, int py)
{
    double h;
    double f;
    ED_AmbientSettings_t *ambientSettings;

    ambientSettings = ED_AmbientSettings(ed);
    h = ambientSettings->hiddenScale;
    f = 0.0;

    if (TileAllowsAmbiance(px, py) && TILEINMAP(x, y))
    {
        if (TileAllowsAmbiance(x, y))
        {
            f = CheckTileLine(px, py, x, y, true) ? 1.0 : h;
        }
        else
        {
            if (!tilemap[x][y])
            {
                f = -h;
            }
            else if ((tilemap[x][y] & 0x80) != 0 && AmbientDoorConnected(x, y))
            {
                f = -ambientSettings->doorVisFactor;
            }
        }
    }

    return f;
}

/*
===============
=
= ConvAmbientVisFactors
=
===============
*/

ED_VisFactorMat_t ConvAmbientVisFactors(ED_VisFactorMat_t tileMat)
{
    int i, c;
    int x, y, r;
    int ineigh;
    double f, h, df, fneigh;
    int n, doorPass;
    ED_VisFactorMat_t convMat;
    lwlib::Mat<int, ED_VISFACTOR_MATRIX_SIZE> passMat;
    ED_AmbientSettings_t *ambientSettings;

    ambientSettings = ED_AmbientSettings(ed);
    r = ambientSettings->radius;
    h = ambientSettings->hiddenScale;
    n = ambientSettings->convPasses;
    df = (1.0 - h) / n;

    convMat = tileMat;

    for (i = 0; i < n; i++)
    {
        for (x = 0; x < 2 * r; x++)
        {
            for (y = 0; y < 2 * r; y++)
            {
                if (mat_elem(convMat, y, x) == h)
                {
                    for (c = 0; c < 9; c++)
                    {
                        fneigh = mat_elem_neigh(tileMat, y, x, c);
                        if (fneigh > 0.0 && fneigh != h)
                        {
                            mat_elem(convMat, y, x) = 1.0 - (df * (i + 1));
                        }
                    }
                }
            }
        }

        tileMat = convMat;
    }

    doorPass = (int)round((1.0 - ambientSettings->doorVisFactor) / df);

    passMat = lwlib::mat_zero<int, ED_VISFACTOR_MATRIX_SIZE>(2 * r, 2 * r);
    for (x = 0; x < 2 * r; x++)
    {
        for (y = 0; y < 2 * r; y++)
        {
            if (mat_elem(tileMat, y, x) == -ambientSettings->doorVisFactor)
            {
                mat_elem(passMat, y, x) = doorPass;
            }
        }
    }

    for (i = 0; i < n; i++)
    {
        for (x = 0; x < 2 * r; x++)
        {
            for (y = 0; y < 2 * r; y++)
            {
                if (mat_elem(convMat, y, x) == -h)
                {
                    for (c = 0; c < 9; c++)
                    {
                        fneigh = mat_elem_neigh(tileMat, y, x, c);
                        ineigh = mat_elem_neigh(passMat, y, x, c);

                        if (fneigh < 0.0 && fneigh != -h)
                        {
                            f = 1.0 - (df * lwlib_MIN(ineigh + 1, n - 1));
                            mat_elem(convMat, y, x) = -f;
                            mat_elem(passMat, y, x) = ineigh + 1;
                        }
                    }
                }
            }
        }

        tileMat = convMat;
    }

    for (x = 0; x < 2 * r; x++)
    {
        for (y = 0; y < 2 * r; y++)
        {
            if (mat_elem(tileMat, y, x) < 0.0)
            {
                mat_elem(tileMat, y, x) = -mat_elem(tileMat, y, x);
            }
        }
    }

    return tileMat;
}


/*
===============
=
= ComputeVisFactorMat
=
===============
*/

static ED_VisFactorMat_t ComputeVisFactorMat(void)
{
    int i;
    int x, y;
    int m, n;
    int r;
    double u, v;
    ED_VisFactorMat_t visFactorMat, tileMat;
    ED_VisFactorMat_t tileVisFactorMats[4];
    ED_AmbientSettings_t *ambientSettings;

    ambientSettings = ED_AmbientSettings(ed);
    r = ambientSettings->radius;
    u = ambientSettings->u;
    v = ambientSettings->v;

    visFactorMat = lwlib::mat_zero<double, ED_VISFACTOR_MATRIX_SIZE>(
        2 * r, 2 * r);

    for (i = 0; i < 4; i++)
    {
        m = ((i + 1) >> 1) & 1;
        n = i >> 1;

        px = ambientSettings->ox + m;
        py = ambientSettings->oy + n;

        tileMat = visFactorMat;

        for (x = player->tilex - r; x < player->tilex + r; x++)
        {
            for (y = player->tiley - r; y < player->tiley + r; y++)
            {
                mat_elem(tileMat, 
                    y - (player->tiley - r),
                    x - (player->tilex - r)) = 
                    ComputeAmbientVisFactor(x, y, px, py);
            }
        }

        tileMat = ConvAmbientVisFactors(tileMat);
        tileVisFactorMats[i] = tileMat;
    }

    visFactorMat = mat_lerp(tileVisFactorMats, u, v);

    return visFactorMat;
}

/*
===============
=
= ComputeAmbientVol
=
= Distance scaling effect for ambient sounds.
=
===============
*/

lwlib::Point2i ComputeAmbientVol(ED_Ambient_t *ambient, 
    const ED_VisFactorMat_t &visFactorMat)
{
    int i;
    int x, y;
    int r;
    double visFactor;
    lwlib::Point2f vols;
    lwlib::Point2i earVols;
    ED_AmbientSettings_t *ambientSettings;

    ambientSettings = ED_AmbientSettings(ed);
    r = ambientSettings->radius;

    vols = lwlib::vec2f_zero();

    for (x = player->tilex - r; x < player->tilex + r; x++)
    {
        for (y = player->tiley - r; y < player->tiley + r; y++)
        {
            visFactor = mat_elem(visFactorMat, 
                y - (player->tiley - r), x - (player->tilex - r));
            vols = vec_add(vols, 
                ComputeAmbientVolTile(ambient, x, y, visFactor));
        }
    }

    for (i = 0; i < 2; i++)
    {
        earVols.v[i] = lwlib_MIN((int)(vols.v[i] * 255), 255);
    }
    return earVols;
}

/*
===============
=
= DoAmbiance
=
= WSJ: play an ambient sound if it's on the
= same floor code as the player
=
===============
*/

void DoAmbiance (void)
{
    int i, j;
    lwlib::Point2i earVolumes;
    int channel;
    voltype vol;
    bool ambianceOn;
    ED_Ambient_t *ambient;
    ASM_SoundPlay_t soundPlay;
    ED_AmbientSettings_t *ambientSettings;
    lwlib::Ray3f ray;
    fixed px, py;
    ED_VisFactorMat_t visFactorMat;

    ambientSettings = ED_AmbientSettings(ed);

    px = player->x - 0x8000;
    py = player->y - 0x8000;
    ambientSettings->ox = lwlib_CLIP(fixedpt_toint(px), 0, MAPSIZE - 2);
    ambientSettings->oy = lwlib_CLIP(fixedpt_toint(py), 0, MAPSIZE - 2);
    ambientSettings->u = fixedpt_toconst(fixedpt_fracpart(px));
    ambientSettings->v = fixedpt_toconst(fixedpt_fracpart(py));

    ray = Player_GetViewRay();
    ray.p[1] = vec_neg(vec_rot90(ray.p[1]));
    ambientSettings->earRay = ray;

    visFactorMat = ComputeVisFactorMat();

    for (i = 0; i < ED_NumAmbients(ed); i++)
    {
        ambient = &ED_Ambients(ed)[i];
        assert(i < 32);

        if (ambient->chan != 0 && 
            !SD_ChannelPlaying(ambient->chan - 1))
        {
            ambient->chan = 0;
        }

        if (ambient->distanceScaled)
        {
            earVolumes = ComputeAmbientVol(ambient, visFactorMat);

            vol = Vol_Zero(true);
            for (j = 0; j < 2; j++)
            {
                Vol_SetStereoVolume(&vol, j, earVolumes.v[j]);
            }

            ambianceOn = !Vol_IsZero(&vol);
        }
        else
        {
            vol = Vol_Mono(SD_MaxVolume());
            ambianceOn = ((ambiance[player->areanumber] | 
                gamestate.ambiance[player->areanumber]) & (1 << i)) != 0;
        }

        if (ambient->looped)
        {
            if (ambianceOn)
            {
                if (ambient->chan == 0)
                {
                    lwlib_Zero(soundPlay);
                    soundPlay.sound = ED_SND(ambient->soundDef);
                    soundPlay.vol = vol;
                    soundPlay.loops = -1;
                    soundPlay.fadeInTime = ambient->fadeInTime;

                    channel = SD_PlaySoundEx(soundPlay);
                    ambient->chan = ((channel != -1) ? channel + 1 : 0);
                }
                if (ambient->distanceScaled && ambient->chan != 0)
                {
                    channel = ambient->chan - 1;
                    SD_SetChannelVol(channel, vol);
                }
            }
            else
            {
                if (ambient->chan != 0)
                {
                    if (ambient->fadeOutTime != 0)
                    {
                        SD_FadeOutChannel(ambient->chan - 1,
                            ambient->fadeOutTime);
                    }
                    else
                    {
                        SD_StopChannel(ambient->chan - 1);
                    }
                    ambient->chan = 0;
                }
            }
        }
        else
        {
            if (ambianceOn)
            {
                ambient->timer += TICS2MILLI(tics);
                if (ambient->timer > ambient->timePeriod)
                {
                    ED_RandomizeAmbientTimePeriod(ed, i);

                    if (ambient->chan == 0)
                    {
                        lwlib_Zero(soundPlay);
                        soundPlay.sound = ED_SND(ambient->soundDef);
                        soundPlay.vol = vol;
                        soundPlay.loops = 0;

                        channel = SD_PlaySoundEx(soundPlay);
                        ambient->chan = ((channel != -1) ? channel + 1 : 0);
                    }
                }
            }
            else
            {
                if (ambient->timer != 0)
                {
                    ED_RandomizeAmbientTimePeriod(ed, i);
                }
            }
        }
    }
}

/*
===============
=
= SetPlayerAmbiance
=
= Sets a specific ambient sound to play on same floor code as the player.
=
===============
*/

void SetPlayerAmbiance(ED_Ambient_t *ambient, bool on)
{
    int i;
    int ambIndex;

    if (ambient == NULL)
    {
        return;
    }

    ambIndex = ambient - ED_Ambients(ed);

    for (i = 0; i < NUMAREAS; i++)
    {
        lwlib_SetBitLValue(gamestate.ambiance[i],
            ambIndex, 0);
    }

    lwlib_SetBitLValue(gamestate.ambiance[player->areanumber],
        ambIndex, on);
}

//==========================================================================


/*
===================
=
= PlayLoop
=
===================
*/

void PlayLoop (void)
{
#if defined(USE_FEATUREFLAGS) && defined(USE_CLOUDSKY)
    if(GetFeatureFlags() & FF_CLOUDSKY)
        InitSky();
#endif

#ifdef USE_SHADING
    InitLevelShadeTable();
#endif

    playstate = ex_stillplaying;
    lasttimecount = GetTimeCount();
    frameon = 0;
    anglefrac = 0;
    facecount = 0;
    memset (buttonstate, 0, sizeof (buttonstate));
    ClearPaletteShifts ();

    if (MousePresent && IN_IsInputGrabbed())
        IN_CenterMouse();         // Clear accumulated mouse movement

    if (demoplayback)
        IN_StartAck ();

    do
    {
		PollControls ();

//
// actor thinking
//
        madenoise = false;
        lastalertedarea = alertedarea;
        alertedarea = 0;

        MoveDoors ();
        MovePWalls ();

        CheckGoalTile ();

        for (ITEROBJ(obj, 0))
        {
            DoActor (obj);
        }

        //
        // WSJ: AMBIENT SOUNDS/LIGHTNING
        //
        DoThunder ();
        Mecha_SetAmbiance();
        DoAmbiance ();

        UpdatePaletteShifts ();

        ThreeDRefresh ();

        gamestate.TimeCount += tics;

        UpdateSoundLoc ();      // JAB
        if (screenfaded)
            VW_FadeIn ();

        CheckKeys ();

//
// debug aids
//
        if (singlestep)
        {
            VW_WaitVBL (singlestep);
            lasttimecount = GetTimeCount();
        }
        if (extravbls)
            VW_WaitVBL (extravbls);

        if (demoplayback)
        {
            if (IN_CheckAck ())
            {
                IN_ClearKeysDown ();
                playstate = ex_abort;
            }
        }
    }
    while (!playstate && !startgame);

    if (playstate != ex_died)
        FinishPaletteShifts ();
}

// mecha

#define MECHA_RUNNINGCHARGE_TICS 100
#define MECHA_RUNNINGCHARGE_DAMAGE 120
#define MECHA_RUNNINGCHARGE_FUEL 35

boolean Mecha_Equipped(void)
{
    return (gamestate.weapon == wp_mecha);
}

boolean Mecha_FuelLow(void)
{
    return gamestate.ammo[ammo_fuel] < MECHA_LOW_FUEL;
}

void T_MechaExit(objtype *ob)
{
    int spawndir;
    int oldtilex, oldtiley;
    controldir_t olddir;
	ED_Spawn_t *spawn;
    const char *spawnName;
    mecha_t *mecha;

    mecha = &gamestate.mecha;

    assert(ob == player);

    oldtilex = player->tilex;
    oldtiley = player->tiley;
    olddir = Player_CardinalDir();

    ExitMountedWeapon(olddir);

    if (player->tilex == oldtilex && 
        player->tiley == oldtiley)
    {
        return;
    }

    spawndir = -1;
    if (mecha->lightWeight)
    {
        spawnName = "en_mechasuit2gempty";
        spawndir = 8;
    }
    else
    {
        spawnName = "en_mechasuit";
        switch (olddir)
        {
        case di_north:
            spawndir = 1;
            break;
        case di_east:
            spawndir = 0;
            break;
        case di_south:
            spawndir = 3;
            break;
        case di_west:
            spawndir = 2;
            break;
        }
    }
    spawn = ED_SpawnByEnemyName(ed, spawnName, spawndir);
    if (spawn == NULL)
    {
        return;
    }

    SpawnActor(oldtilex, oldtiley, spawn->actor);

    if (!mecha->lightWeight)
    {
        newobj->hitpoints = gamestate.armor;
        gamestate.armor = mecha->playerarmor;
    }
    DrawWeapon();

    if (!mecha->lightWeight)
    {
        if (newobj->hitpoints < MECHA_HP_NOARMS(mecha->starthp))
        {
            NewState(newobj, ED_STATE("s_mechasuitnoarms"));
        }
        else if (newobj->hitpoints < MECHA_HP_ONEARM(mecha->starthp))
        {
            NewState(newobj, ED_STATE("s_mechasuitonearm"));
        }
    }

    lwlib_Zero(*mecha);
}

void T_MechaEject(objtype *ob)
{
    mecha_t *mecha;

    assert(ob == player);

    mecha = &gamestate.mecha;
    assert(!mecha->lightWeight);

    GetMessage ("Your Mecha was destroyed!");
    SD_PlaySound (ED_SND("EJECTMECHASND"));

    buttonstate[bt_attack] = false;
    buttonheld[bt_attack] = false;
    gamestate.attackframe = 0;
    gamestate.weaponframe = 0;
    gamestate.weapon = gamestate.chosenweapon;
    gamestate.armor = mecha->playerarmor;
    DrawWeapon ();
    DrawAmmo ();

    SpawnNewObj (player->tilex, player->tiley, ED_STATE("s_mechasuitdie4"));
    A_RemoveBlocking(newobj);
    actorat[newobj->tilex][newobj->tiley] = NULL;
    newobj->flags |= FL_NONMARK | FL_NEVERMARK;

    lwlib_Zero(*mecha);
}

bool Mecha_HasNoArms(void)
{
    mecha_t *mecha;

    assert(Mecha_Equipped());
    mecha = &gamestate.mecha;

    return !mecha->lightWeight && gamestate.armor < 
        MECHA_HP_NOARMS(mecha->starthp);
}

void Mecha_SetAmbiance(void)
{
    mecha_t *mecha;

    mecha = &gamestate.mecha;

    SetPlayerAmbiance(ED_AMB("MECHAENGINESND"), Mecha_Equipped() && 
        !Mecha_FuelLow() && gamestate.ammo[ammo_fuel] > 0);
    SetPlayerAmbiance(ED_AMB("MECHAENGINELOWSND"), Mecha_Equipped() && 
        Mecha_FuelLow() && gamestate.ammo[ammo_fuel] > 0);

    mecha->rotateTics -= tics;
    if (mecha->rotateTics < 0)
    {
        mecha->rotateTics = 0;
    }

    SetPlayerAmbiance(ED_AMB("MECHAROTATESND"), Mecha_Equipped() && 
        mecha->rotating);
}

void Mecha_NotifyRotating(bool rotating)
{
    int volume;
    int fadeTics;
    mecha_t *mecha;
    ED_Ambient_t *ambient;

    mecha = &gamestate.mecha;
    ambient = ED_AMB("MECHAROTATESND");
    if (ambient == NULL)
    {
        return;
    }

    fadeTics = MILLI2TICS(ambient->fadeInTime);

    if (rotating && abs(controlx) < BASEMOVE)
    {
        rotating = false;
    }

    if (!rotating && mecha->rotating)
    {
        mecha->stopTics += tics;
        if (mecha->stopTics < fadeTics / 2)
        {
            rotating = true;
        }
        else
        {
            mecha->stopTics = 0;
        }
    }
    else
    {
        mecha->stopTics = 0;
    }

    if (rotating)
    {
        if (!mecha->rotating)
        {
            mecha->rotateTics = fadeTics;
        }
    }
    else
    {
        if (mecha->rotating)
        {
            volume = SD_MaxVolume();
            if (mecha->rotateTics > 0)
            {
                volume = (fadeTics - mecha->rotateTics) * 
                    SD_MaxVolume() / fadeTics;
            }

            SD_PlaySoundVolume(ED_SND("MECHAROTATESTOPSND"), volume);
        }
        mecha->rotateTics = 0;
    }

    mecha->rotating = rotating;
}

void Mecha_StartRunningCharge(void)
{
    mecha_t *mecha;

    mecha = &gamestate.mecha;

    if (mecha->runningCharge.active)
    {
        return;
    }

    assert(Mecha_Equipped());

    if (CURAMMO <= MECHA_RUNNINGCHARGE_FUEL)
    {
        return;
    }

    ConsumeAmmo(MECHA_RUNNINGCHARGE_FUEL);

    mecha = &gamestate.mecha;

    lwlib_Zero(mecha->runningCharge);
    mecha->runningCharge.active = true;
    mecha->runningCharge.ticsPassed = 0;
    mecha->runningCharge.angle = player->fineangle;

    SD_PlaySound (ED_SND("MECHARUNNINGCHARGESND"));
}

void Mecha_StopRunningCharge(void)
{
    mecha_t *mecha;

    assert(Mecha_Equipped());

    mecha = &gamestate.mecha;
    lwlib_Zero(mecha->runningCharge);
}

void Mecha_ControlRunningCharge(void)
{
    mecha_t *mecha;

    if (!Mecha_Equipped())
    {
        return;
    }

    mecha = &gamestate.mecha;

    mecha->runningCharge.ticsPassed += tics;
    if (mecha->runningCharge.ticsPassed >= MECHA_RUNNINGCHARGE_TICS)
    {
        Mecha_StopRunningCharge();
    }
}

void Mecha_NotifyHitSolid(void)
{
    mecha_t *mecha;

    mecha = &gamestate.mecha;

    if (!(Mecha_Equipped() && mecha->runningCharge.active))
    {
        return;
    }

    //Mecha_StopRunningCharge();
}

void Mecha_NotifyHitObj(objtype *ob)
{
    mecha_t *mecha;

    mecha = &gamestate.mecha;

    if (!(Mecha_Equipped() && mecha->runningCharge.active))
    {
        return;
    }

    DamageActor(ob, player, MECHA_RUNNINGCHARGE_DAMAGE, dam_rocket);
    if (ob->hitpoints > 0)
    {
        Mecha_StopRunningCharge();
    }
    else
    {
        A_RemoveBlocking(ob);
    }
}

short Mecha_GetViewingAngle(void)
{
    return Mecha_GetViewingFineAngle() * ANGLES / FINEANGLES;
}

short Mecha_GetViewingFineAngle(void)
{
    short fineangle;
    mecha_t *mecha;

    fineangle = player->fineangle;

    if (Mecha_Equipped())
    {
        mecha = &gamestate.mecha;
        fineangle += mecha->motionTurnAngle;
    }

    while (fineangle >= FINEANGLES)
    {
        fineangle -= FINEANGLES;
    }
    while (fineangle < 0)
    {
        fineangle += FINEANGLES;
    }

    return fineangle;
}

bool Mecha_RunningChargeActive(void)
{
    mecha_t *mecha;
    mecha = &gamestate.mecha;
    return Mecha_Equipped() && mecha->runningCharge.active;
}

int Mecha_GetStatusBarPic(void)
{
    return Mecha_Equipped() ? MECHASTATUSBARPIC : STATUSBARPIC;
}

// player

controldir_t Player_CardinalDir(void)
{
    controldir_t dir;

    if (player->angle < ANGLES/8 || player->angle > 7*ANGLES/8)
    {
        dir = di_east;
    }
    else if (player->angle < 3*ANGLES/8)
    {
        dir = di_north;
    }
    else if (player->angle < 5*ANGLES/8)
    {
        dir = di_west;
    }
    else
    {
        dir = di_south;
    }

    return dir;
}

void Player_GetFacing(Player_Facing_t *facing)
{
    memset(facing, 0, sizeof(Player_Facing_t));

    facing->dir = Player_CardinalDir();
    if (facing->dir == di_east)
    {
        facing->checkx = player->tilex + 1;
        facing->checky = player->tiley;
        facing->elevatorok = true;
    }
    else if (facing->dir == di_north)
    {
        facing->checkx = player->tilex;
        facing->checky = player->tiley-1;
        facing->elevatorok = false;
    }
    else if (facing->dir == di_west)
    {
        facing->checkx = player->tilex - 1;
        facing->checky = player->tiley;
        facing->elevatorok = true;
    }
    else
    {
        facing->checkx = player->tilex;
        facing->checky = player->tiley + 1;
        facing->elevatorok = false;
    }
}

fixed Player_CamGetX(void)
{
    tremor_t *tremor;
    tremor = &gamestate.tremor;
    return OB_GetX(player) + gamestate.camx + tremor->camxOff;
}

fixed Player_CamGetY(void)
{
    tremor_t *tremor;
    tremor = &gamestate.tremor;
    return OB_GetY(player) + gamestate.camy + tremor->camyOff;
}

int Player_CamGetTileX(void)
{
    return Player_CamGetX() >> TILESHIFT;
}

int Player_CamGetTileY(void)
{
    return Player_CamGetY() >> TILESHIFT;
}

const lwlib::Ray3f Player_GetViewRay(void)
{
    return OB_GetViewRay(player);
}
