// WL_RESURRECT.CPP

#include <stdio.h>
#include <math.h>
#include "wl_def.h"
#include "lw.h"
#if defined _WIN32
#pragma hdrstop
#endif

/*
===================================================================

                           RESURRECTION

===================================================================
*/

bool OB_CanResurrect(objtype *ob)
{
    ED_Resurrection_t *resurrection;
    resurrection = ob->enemy->resurrection;
    return resurrection != NULL &&
        resurrection->enterBodyState[0] != '\0';
}

bool OB_CanTargetObjResurrect(objtype *ob, objtype *check)
{
    ED_Resurrection_t *resurrection;
    resurrection = check->enemy->resurrection;
    return resurrection != NULL && 
        resurrection->undeadName[0] != '\0' && 
        check->deathtyp == death_normal && 
        (check->flags & FL_BLOCK) == 0;
}

void A_ReachedBody(objtype *ob)
{
    ED_Resurrection_t *resurrection;

    resurrection = ob->enemy->resurrection;
    PlayActorVoice(ED_SND(
        resurrection->enterBodySound), ob);

    NewState(ob, ED_STATE(resurrection->enterBodyState));
}

void A_ResurrectBody(objtype *ob)
{
    int goalObjIndex;
    objtype *goalObj, *deadObj, *undeadObj;
    ED_Enemy_t *deadEnemy, *undeadEnemy;
    ED_State_t *skinState;

    goalObjIndex = ob->objIndexArr[oit_goaltarget];
    assert(goalObjIndex != 0);

    goalObj = &objlist[goalObjIndex];
    deadObj = goalObj;
    deadEnemy = deadObj->enemy;

    assert(deadEnemy->resurrection != NULL && 
        deadEnemy->resurrection->undeadName[0] != '\0');

    undeadEnemy = ED_Enemy(ed, 
        deadEnemy->resurrection->undeadName);
    assert(undeadEnemy->resurrection != NULL);

    SpawnNewObj(ob->tilex, ob->tiley, 
        ED_STATE(undeadEnemy->resurrection->resurrectState));
    undeadObj = newobj;
    OB_SetEnemy(undeadObj, undeadEnemy);

    undeadObj->hitpoints = undeadObj->starthitpoints = 
        undeadEnemy->health;
    undeadObj->killpoints = ob->killpoints;

    undeadObj->dir = nodir;
    undeadObj->flags |= FL_BLOCK;
    undeadObj->x = ob->x;
    undeadObj->y = ob->y;

    T_Aim(undeadObj);

    skinState = ED_STATE(deadEnemy->resurrection->skinState);
    if (skinState != NULL)
    {
        NewState(deadObj, skinState);
    }
    else
    {
        OB_SetState(deadObj, NULL);      // mark for removal
    }

    PlayActorVoice(ED_SND(
        undeadEnemy->resurrection->resurrectSound), undeadObj);

    if (CheckPlaysMusic(undeadObj))
    {
        UpdateBossMusic(undeadObj, true);
    }
}

bool OB_CheckSummonBodyInSight(objtype *ob, int &deadObjNum)
{
    objtype *check;
    const int maxDeadObjs = 16;
    int deadObjNums[maxDeadObjs];

    for (int i = 0; i < maxDeadObjs; i++)
    {
        deadObjNums[i] = -1;
    }

    int deadObjCount = 0;

    for (ITEROBJ(check, 1))
    {
        if (!OB_CanTargetObjResurrect(ob, check))
        {
            // body is not resurrectable
            continue;
        }

        if (check->objIndexArr[oit_goalseeker] != 0)
        {
            // this object is already targeted by another actor
            continue;
        }

        if (!CheckObjSight(ob, player) ||
            !CheckObjSight(check, player) ||
            !spotvis[ob->tilex][ob->tiley] ||
            !spotvis[check->tilex][check->tiley])
        {
            // dead body and summoner must be visible by player
            continue;
        }

        if ((check->flags & FL_SPAWNEDGHOST) != 0)
        {
            // enemy already spawned a ghost
            continue;
        }

        if (deadObjCount < maxDeadObjs)
        {
            deadObjNums[deadObjCount++] = check - objlist;
        }
    }

    if (deadObjCount > 0)
    {
        deadObjNum = deadObjNums[US_RndT_Range(0, deadObjCount - 1)];
    }

    return deadObjCount > 0;
}

void A_SummonGhostFromDead(objtype *ob)
{
    objtype *shooterObj = NULL;

    int deadObjNum = -1;
    if (OB_CheckSummonBodyInSight(ob, deadObjNum))
    {
        assert(deadObjNum >= 0);
        objtype *deadObj = objlist + deadObjNum;

        if ((deadObj->flags & FL_SPAWNEDGHOST) == 0)
        {
            shooterObj = deadObj;
            deadObj->flags |= FL_SPAWNEDGHOST;
        }
        else
        {
            shooterObj = ob;
        }
    }
    else
    {
        shooterObj = ob;
    }

    if (shooterObj != NULL)
    {
        const short angleoffset = US_RndT_Range(0, 5);

        T_Aim(shooterObj);
        ProjFire(shooterObj,angleoffset,
            ED_Proj(ed, (shooterObj == ob ? "deathskull" : "summonskull")),
            NULL,8192,0);
    }
}

void A_ResetSpawnedGhostFlag(objtype *ob)
{
    objtype *check;

    for (ITEROBJ(check, 1))
    {
        check->flags &= ~FL_SPAWNEDGHOST;
    }
}
